from .base import BaseValidatedModel
from pydantic import Field, model_validator
from typing import Optional, Union, Dict
    
# Base class for image/video
class MediaBaseModel(BaseValidatedModel):
    inner_mediaurl: str = Field(..., min_length=1, strict=True)
    media_size: Optional[int] = Field(default=None, strict=True)

    media_data: Optional[bytes] = Field(default=None, strict=True) # Raw bytes of the image/video
    media_type: Optional[str] = Field(default=None, strict=True) # MIME type of the image or video cover (e.g., "image/png")

    upload_url: Optional[str] = Field(default=None, strict=True)
    upload_headers: Optional[Dict[str, Union[str, int]]] = Field(default=None, strict=True)
    upload_data: Optional[bytes] = Field(default=None, strict=True)
    fill_text: Optional[str] = Field(default=None, strict=True)

class VideoInfo(MediaBaseModel):
    @model_validator(mode='after')
    def check_video_fields(self):
        if self.media_size is None or self.media_size <= 0:
            raise ValueError("Video must provide media_size > 0")
        return self

class PhotoInfo(MediaBaseModel):
    pass

class MediaInfo(BaseValidatedModel):
    content_type: int # 0 = video, 1 = image
    video_info: Optional[VideoInfo] = None
    photo_info: Optional[PhotoInfo] = None
    media_height: Optional[int] = 0 # Optional height of the uploaded image
    media_width: Optional[int] = 0 # Optional width of the uploaded image

    @model_validator(mode="after")
    def validate_media(self):
        if self.content_type not in [0, 1]:
            raise ValueError(f"Unknown content_type: {self.content_type}")
        if self.content_type == 0 and not self.video_info:
            raise ValueError("video_info must be provided for content_type=0 (video)")
        if self.content_type == 1 and not self.photo_info:
            raise ValueError("photo_info must be provided for content_type=1 (image)")
        return self



import asyncio, json
from ..core.downloader.internet import Request, HttpRequest, WebSocketRequest
from ..utils import do_sha1
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ..spiders import Spider
    from ..settings import SettingsInfo
    from ..cpy.cpy_resources.bloom.fallback import BloomFilterPy

class BaseFingerprint:
    def __init__(self, settings: "SettingsInfo"=None, **kwargs):
        self.settings = settings
        self.include_headers = self.settings.INCLUDE_HEADERS
        self.kwargs = kwargs

    def create_bytes(self, request: Request) -> bytes:
        parts = []
        if not isinstance(self.include_headers, list):
            raise ValueError("INCLUDE_HEADERS in settings is not list.")
        include_headers = {}
        for header_key in self.include_headers:
            has_header_key =  request.find_header_key(key=header_key)
            if has_header_key:
                include_headers[has_header_key.lower()] = request.headers[has_header_key]
        parts.append(f'{request.url}|{json.dumps(include_headers, separators=(",", ":"), sort_keys=True)}'.encode('latin-1'))
        if isinstance(request, HttpRequest):
            parts.append(f'{request.method}|'.encode('latin-1'))
            if isinstance(request.data, bytes):
                parts.append(request.data)
            elif isinstance(request.data, dict):
                parts.append(json.dumps(request.data, separators=(",", ":"), sort_keys=True).encode('latin-1'))
        elif isinstance(request, WebSocketRequest):
            for msg in request.send_message:
                parts.append(msg.data)
        return b''.join(parts)

    def get_fingerprint(self, request: Request) -> str:
        return do_sha1(self.create_bytes(request))
    
class MemoryDupeFilter(BaseFingerprint):
    def __init__(self, settings: "SettingsInfo"=None, **kwargs):
        super().__init__(settings=settings, **kwargs)
        self.new_seen: set[str] = set() # Requests marked as seen but not yet sent
        self.sent_seen: set[str] = set() # Requests that have been seen and already sent
        self.lock = asyncio.Lock()
        
    async def request_seen(self, request: "Request"=None, **kwargs):
        fingerprint = self.get_fingerprint(request=request)
        is_seen = fingerprint in self.new_seen
        # If the request fingerprint is already in new_seen (i.e., seen before), return True.
        # Otherwise, add the fingerprint to new_seen and check if it is present in sent_seen,
        # which indicates the request has already been dispatched.
        if is_seen:
            return is_seen
        self.new_seen.add(fingerprint)
        return fingerprint in self.sent_seen
    
    async def mark_sent(self, request: "Request", spider: "Spider", **kwargs):
        if not request.dont_filter:
            async with self.lock:
                self.sent_seen.add(self.get_fingerprint(request=request))
    
class BloomDupeFilter(BaseFingerprint):
    def __init__(self, settings: "SettingsInfo"=None, **kwargs):
        super().__init__(settings=settings, **kwargs)
        import bloom
        self.new_seen: "BloomFilterPy" = bloom.BloomFilter(size=self.settings.BLOOM_INFO.SIZE, expected=self.settings.BLOOM_INFO.EXPECTED, hash_count=self.settings.BLOOM_INFO.HASH_COUNT)
        self.sent_seen: "BloomFilterPy" = bloom.BloomFilter(size=self.settings.BLOOM_INFO.SIZE, expected=self.settings.BLOOM_INFO.EXPECTED, hash_count=self.settings.BLOOM_INFO.HASH_COUNT)
        self.lock = asyncio.Lock()

    async def request_seen(self, request: "Request"=None, **kwargs):
        origin_fp_bytes = self.create_bytes(request=request)
        is_seen = self.new_seen.exists(origin_fp_bytes)
        # If the request fingerprint is already in new_seen (i.e., seen before), return True.
        # Otherwise, add the fingerprint to new_seen and check if it is present in sent_seen,
        # which indicates the request has already been dispatched.
        if is_seen:
            return is_seen
        self.new_seen.add(origin_fp_bytes)
        return self.sent_seen.exists(origin_fp_bytes)

    async def mark_sent(self, request: "Request", spider: "Spider", **kwargs):
        if not request.dont_filter:
            async with self.lock:
                self.sent_seen.add(self.create_bytes(request=request))


try:
    import lz4.frame
    import lz4.block
except ImportError as e:
    raise ImportError(
        "Missing lz4 dependencies. "
        "Please install: pip install lz4"
    ) from e

class Lz4Factory(object):
    @staticmethod
    def frame_encode(data: bytes, compression_level: int = 0) -> bytes:
        return lz4.frame.compress(data, compression_level=compression_level)

    @staticmethod
    def frame_decode(data: bytes) -> bytes:
        return lz4.frame.decompress(data)

    @staticmethod
    def block_encode(data: bytes, compression: bool = True) -> bytes:
        return lz4.block.compress(data) if compression else data

    @staticmethod
    def block_decode(data: bytes, uncompressed_size: int = None) -> bytes:
        if uncompressed_size:
            return lz4.block.decompress(data, uncompressed_size=uncompressed_size)

        if not uncompressed_size:
            last_exc = None
            uncompressed_size_list = [len(data)*3, len(data)*5, int(1024e3)]
            for uncompressed_size in uncompressed_size_list:
                try:
                    return lz4.block.decompress(data, uncompressed_size=uncompressed_size)
                except Exception as e:
                    last_exc = e
            raise ValueError("Unable to lz4-decompress.") from last_exc
    
__all__ = [
    "Lz4Factory"
]


import asyncio
from typing import Union
from ..spiders import Spider
from ..item import Item
from ..hooks import interceptors_hooks
from ..core.downloader.internet import *
from typing import TYPE_CHECKING, Dict
if TYPE_CHECKING:
    from ..crawler import Crawler
    from ..hooks.interceptors import InterceptorsHooks
    from ..settings import SettingsInfo
    from ..mq.kafka import KafkaManager

class BaseInterceptor:
    def __init__(
        self, 
        stop_event: asyncio.Event=None, 
        settings: "SettingsInfo"=None, 
        kafkaManager: "KafkaManager"=None,
        **kwargs
    ):
        self.stop_event = stop_event
        self.settings = settings
        self.kafkaManager = kafkaManager
        self.kwargs = kwargs

    @classmethod
    def from_crawler(cls, crawler: "Crawler"):
        return cls(
            stop_event=crawler.stop_event,
            settings=crawler.settings,
            kafkaManager=crawler.kafkaManager,
        )

class DownloadInterceptor(BaseInterceptor):
    RequestType = Union[HttpRequest, WebSocketRequest]
    ResponseType = Union[HttpResponse, WebSocketResponse]

    async def request_intercept(self, request: RequestType, spider: Spider):
        return None

    async def response_intercept(self, request: RequestType, response: ResponseType, spider: Spider):
        return response
    
    async def exception_intercept(self, request: RequestType, exception: BaseException, spider: Spider):
        return exception
    
class SpiderInterceptor(BaseInterceptor):
    ResponseType = Union[HttpResponse, WebSocketResponse]
    ResultType = Union[Request, Item, Dict, None]
    async def process_spider_input(self, response: ResponseType, spider: Spider):
        return None

    async def process_spider_output(self, response: ResponseType, result: ResultType, spider: Spider):
        return result
    
    async def process_spider_exception(self, response: ResponseType, exception: BaseException, spider: Spider):
        return None

class _InnerSpiderInterceptor(SpiderInterceptor):
    def __init__(self,  
        stop_event: asyncio.Event=None, 
        settings: "SettingsInfo"=None, 
        hooks: "InterceptorsHooks"=None, 
        sessions_lock: asyncio.Lock=None, 
        kafkaManager: "KafkaManager"=None,
        **kwargs
    ):
        super().__init__(stop_event=stop_event, settings=settings, kafkaManager=kafkaManager, **kwargs)
        self.hooks = hooks
        self.sessions_lock = sessions_lock

    @classmethod
    def from_crawler(cls, crawler: "Crawler"):
        return cls(
            stop_event=crawler.stop_event,
            settings=crawler.settings,
            hooks=interceptors_hooks(crawler),
            sessions_lock=crawler.sessions_lock,
            kafkaManager=crawler.kafkaManager,
        )


import shutil, os
from pathlib import Path
from typing import List

def copytree_merge(src: Path, dst: Path):
    src = Path(src)
    dst = Path(dst)
    if not src.is_dir():
        raise ValueError(f"not dir: {src}")

    if not dst.exists():
        os.makedirs(dst)

    for item in src.iterdir():
        s = src / item.name
        d = dst / item.name
        if s.is_dir():
            copytree_merge(s, d)
        else:
            shutil.copy2(s, d)

def run(use_redis: bool, use_rabbitmq: bool, use_kafka: bool):
    base = Path(__file__).parent.parent # scrapy_cffi
    template_dir = base / "templates"
    target: Path = Path.cwd() / "demo"

    from .base import find_project_root
    from .genspider import check_config
    project_path = find_project_root(is_demo=True)
    check_config(project_path, use_redis=use_redis,  use_rabbitmq=use_rabbitmq, use_kafka=use_kafka)

    settings_path = target / "settings.py"
    settings_code = settings_path.read_text(encoding='utf-8')
    settings_code = settings_code.replace('# settings.EXTENSIONS_PATH', 'settings.EXTENSIONS_PATH')
    settings_code = settings_code.replace('# settings.ITEM_PIPELINES_PATH', 'settings.ITEM_PIPELINES_PATH')
    settings_code = settings_code.replace('# "interceptors.CustomDownloadInterceptor1"', '"interceptors.CustomDownloadInterceptor1"')
    settings_code = settings_code.replace('# "interceptors.CustomDownloadInterceptor2"', '"interceptors.CustomDownloadInterceptor2"')
    settings_path.write_text(settings_code, encoding='utf-8')

    spider_dir = target / "spiders"
    demo_spiders_dir = template_dir / "demo_spider"
    
    # demo_server
    copytree_merge(template_dir / "server", target)
    readme_path = target / "readme.txt"
    readme_code = readme_path.read_text(encoding='utf-8')
    if use_rabbitmq:
        readme_code += (
            "\n"
            "3. Push a task into RabbitMQ.\n"
            "\tYou don’t need to modify anything — the demo project works out of the box.\n"
            "\tFor pushing tasks, simply refer to the ready-to-use test file in the open-source repo:\n"
            "\thttps://github.com/aFunnyStrange/scrapy_cffi/tree/main/tests/test_rabbitmq.py"
        )
    elif use_redis:
        readme_code = readme_code + '\n3.redis-cli\n4.RPUSH customRedisSpider_test http://127.0.0.1:8002\r\n'
    readme_path.write_text(readme_code, encoding='utf-8')

    if use_rabbitmq or use_redis:
        demo_spider_files = ["customRedisSpider", "studentSpider"]
        for demo_spider in demo_spider_files:
            demo_spider_path = demo_spiders_dir / f'{demo_spider}.py'
            target_spider_path = spider_dir / f'{demo_spider}.py'
            demo_spider_code = demo_spider_path.read_text(encoding='utf-8')
            target_spider_path.parent.mkdir(parents=True, exist_ok=True)
            if use_rabbitmq:
                demo_spider_code = demo_spider_code.replace("from scrapy_cffi.spiders.redis import RedisSpider", "from scrapy_cffi.spiders.rabbitmq import RabbitmqSpider")
                demo_spider_code = demo_spider_code.replace("(RedisSpider)", "(RabbitmqSpider)")
                demo_spider_code = demo_spider_code.replace('redis_key = "customRedisSpider_test"', 'rabbitmq_queue = "scrapy_cffi"')
                demo_spider_code = demo_spider_code.replace("'RPUSH customRedisSpider_test http://127.0.0.1:8002'", "")
            target_spider_path.write_text(demo_spider_code, encoding='utf-8')

        update_spiders_path(
            project_path=target, 
            demo_spiders_dir=demo_spiders_dir, 
            demo_spider_files=demo_spider_files, 
            spider_dir=spider_dir, 
            use_redis=use_redis,
            use_rabbitmq=use_rabbitmq
        )

        spiders_dir = target
        runner_path = spiders_dir / "runner.py"
        runner_code = runner_path.read_text(encoding='utf-8')
        runner_code = runner_code.replace('crawler, engine_task = await advance_main()', '# crawler, engine_task = await advance_main()')
        runner_code = runner_code.replace('# crawler, engine_task = await advance_main_all()', 'crawler, engine_task = await advance_main_all()')
        runner_code = runner_code.replace('import threading', '# import threading')
        runner_code = runner_code.replace('t = threading.Thread(', '# t = threading.Thread(')
        runner_code = runner_code.replace('t.start()', '# t.start()')
        runner_code = runner_code.replace('t.join()', '# t.join()')
        runner_code = runner_code.replace(' main()', ' # main()')
        runner_path.write_text(runner_code, encoding='utf-8')
    else:
        spider_dir.mkdir(parents=True, exist_ok=True)
        demo_spider_files = ["customSpider", "studentSpider"]
        for demo_spider in demo_spider_files:
            demo_spider_path = demo_spiders_dir / f'{demo_spider}.py'
            target_spider_path = spider_dir / f'{demo_spider}.py'
            demo_spider_code = demo_spider_path.read_text(encoding='utf-8')
            target_spider_path.parent.mkdir(parents=True, exist_ok=True)
            target_spider_path.write_text(demo_spider_code, encoding='utf-8')
        update_spiders_path(project_path=target, demo_spiders_dir=demo_spiders_dir, demo_spider_files=demo_spider_files, spider_dir=spider_dir, use_redis=use_redis, use_rabbitmq=use_rabbitmq)

    print(f"Project 'demo' created.")

def update_spiders_path(project_path: Path, demo_spiders_dir: Path, demo_spider_files: List, spider_dir: Path, use_redis: bool, use_rabbitmq: bool):
    for spider_name in demo_spider_files:
        # spider_path = demo_spiders_dir / f"{spider_name}.py"
        # spider_code = spider_path.read_text('utf-8')

        # write_path = spider_dir / f"{spider_name}.py"
        # write_path.write_text(spider_code, encoding='utf-8')
        cls_name = spider_name[0].upper() + spider_name[1:] if spider_name else spider_name
        # update __init__.py
        from .genspider import update_spiders_init
        update_spiders_init(project_path=project_path, class_name=cls_name, spider_name=spider_name)


__version__ = "0.1.0"

from .crawler import run_spider, run_all_spiders, run_spider_sync, run_all_spiders_sync, cleanup_loop
from .utils import load_settings_with_path, init_logger

__all__ = [
    "run_spider",
    "run_all_spiders",
    "run_spider_sync",
    "run_all_spiders_sync",
    "load_settings_with_path",
    "init_logger"
]


from pydantic import Field, model_validator
from enum import Enum
from typing import Optional, Union, List
from .base import StrictValidatedModel

class MQMode(str, Enum):
    SINGLE = "single"
    CLUSTER = "cluster"

class BaseMQInfo(StrictValidatedModel):
    DRIVER: Optional[str] = "amqp"
    URL: Optional[str] = None
    HOST: Optional[str] = None
    PORT: Optional[Union[str, int]] = None
    USERNAME: Optional[str] = None
    PASSWORD: Optional[str] = None
    MODE: Union[MQMode, str] = MQMode.SINGLE
    CLUSTER_NODES: Optional[List[str]] = Field(default_factory=list)

    @model_validator(mode="after")
    def assemble_url(self) -> "BaseMQInfo":
        if not self.URL and self.HOST and self.PORT:
            auth_part = ""
            if self.USERNAME and self.PASSWORD:
                auth_part = f"{self.USERNAME}:{self.PASSWORD}@"
            elif self.PASSWORD:
                auth_part = f":{self.PASSWORD}@"
            self.URL = f"{self.DRIVER}://{auth_part}{self.HOST}:{self.PORT}"
        return self

    @property
    def resolved_url(self) -> Union[str, List[str], None]:
        if self.MODE == MQMode.SINGLE:
            return self.URL
        elif self.MODE == MQMode.CLUSTER:
            return self.CLUSTER_NODES
        return None

class RabbitMQInfo(BaseMQInfo):
    VHOST: Optional[str] = "/"
    EXCHANGE_NAME: Optional[str] = "scrapy_cffi"
    EXCHANGE_TYPE: Optional[str] = "direct"
    PREFETCH_COUNT: Optional[int] = 10
    DONT_FILTER: Optional[bool] = False

    @model_validator(mode="after")
    def assemble_url(self) -> "RabbitMQInfo":
        super().assemble_url()
        if self.URL and self.VHOST:
            vhost_part = f"/{self.VHOST.strip('/')}"
            if not self.URL.endswith(vhost_part):
                self.URL = f"{self.URL}{vhost_part}"
        return self

class KafkaInfo(BaseMQInfo):
    CONSUMER_GROUP: Optional[str] = "scrapy_cffi"
    PERSISTENT_TIME: Optional[int] = 7*24*60*60*1000

__all__ = [
    "RabbitMQInfo",
    "KafkaInfo",
]


from .base import StrictValidatedModel
from ..utils import load_object
from typing import List, Type, Union, Dict, Any
from pydantic import Field

class ComponentInfo(StrictValidatedModel):
    value: List[Type] = Field(default_factory=list)

    @classmethod
    def from_raw(cls, raw: Union[Dict[str, Any], List[Any], str, type], field_name="") -> "ComponentInfo":
        # dict -> sort by value and use key as path or class
        if isinstance(raw, dict):
            try:
                sorted_items = sorted(raw.items(), key=lambda item: item[1])
                items = [cls._load(k, field_name) for k, _ in sorted_items]
                return cls(value=items)
            except Exception:
                raise ValueError(f"{field_name}: dict values must be sortable")
        # list -> supports str or class
        elif isinstance(raw, list):
            if all(isinstance(i, (str, type)) for i in raw):
                return cls(value=[cls._load(i, field_name) for i in raw])
            raise ValueError(f"{field_name}: list elements must be str or class")
        # single str or class
        elif isinstance(raw, (str, type)):
            return cls(value=[cls._load(raw, field_name)])
        elif raw is None:
            cls(value=[])
        else:
            raise ValueError(f"{field_name}: must be dict, list[str|class], str, or class")

    @staticmethod
    def _load(obj: Union[str, type], field_name: str) -> Type:
        if isinstance(obj, type):
            return obj
        try:
            return load_object(obj)
        except Exception as e:
            raise ValueError(f"{field_name}: failed to load class from path '{obj}': {e}")
        
__all__ = [
    "ComponentInfo"
]


import os
import multiprocessing, threading
try:
    from scrapy.utils.project import get_project_settings
    from scrapy.spiderloader import SpiderLoader
    from scrapy.cmdline import execute
    from twisted.internet import reactor
    from scrapy.crawler import CrawlerRunner
    from scrapy.utils.log import configure_logging
except ImportError as e:
    raise ImportError(
        "Missing scrapy dependencies. "
        "Please install: pip install scrapy"
    ) from e

from typing import List, TYPE_CHECKING
if TYPE_CHECKING:
    from scrapy.settings import Settings

class ScrapyRunner:
    def __init__(self):
        self.settings = get_project_settings()

    def get_all_spider_names(self) -> List[str]:
        spider_loader = SpiderLoader.from_settings(self.settings)
        spiders = spider_loader.list()
        print(f"There are {len(spiders)} spiders: {spiders}")
        return spiders

    def run_all_spiders(self, spiders: List[str]=None) -> None:
        spiders = spiders or self.get_all_spider_names()
        for spider_name in spiders:
            p = multiprocessing.Process(target=self.run_spider, args=(spider_name,), daemon=True)
            p.start()
            print(f"Start spider：{spider_name}，pid={p.pid}")

    def run_spider(self, spider_name: str) -> None:
        os.environ.setdefault('SCRAPY_SETTINGS_MODULE', 'ins_collect.settings')
        execute(["scrapy", "crawl", spider_name])

def run_spider_in_process(settings: "Settings", spider_name: str) -> None:
    runner = CrawlerRunner(settings)
    d = runner.crawl(spider_name)
    d.addBoth(lambda _: reactor.stop())
    reactor.run()

class InlineScrapyRunner:
    """Run Scrapy spiders in the current process in a non-blocking way using CrawlerRunner."""

    def __init__(self, settings_module: str = "myproject.settings"):
        os.environ.setdefault("SCRAPY_SETTINGS_MODULE", settings_module)
        self.settings: "Settings" = get_project_settings()
        configure_logging()
        self.runner = CrawlerRunner(self.settings)

    def get_all_spider_names(self) -> List[str]:
        spider_loader = SpiderLoader.from_settings(self.settings)
        spiders = spider_loader.list()
        print(f"There are {len(spiders)} spiders: {spiders}")
        return spiders
    
    def run_all_spiders(self, spiders: List[str]=None, use_process: bool = True) -> None:
        spiders = spiders or self.get_all_spider_names()
        for spider_name in spiders:
            self.run_spider(spider_name=spider_name, use_process=use_process)

    def run_spider(self, spider_name: str, use_process: bool = True) -> None:
        if use_process:
            p = multiprocessing.Process(
                target=run_spider_in_process,
                args=(self.settings, spider_name),
                daemon=True
            )
            p.start()
            print(f"[Process mode] Start spider {spider_name}, pid={p.pid}")
        else:
            self.runner.crawl(spider_name)
            if not reactor.running:
                threading.Thread(
                    target=reactor.run,
                    kwargs={"installSignalHandlers": False},
                    daemon=True
                ).start()

__all__ = [
    "ScrapyRunner",
    "InlineScrapyRunner"
]

if __name__ == "__main__":
    os.chdir(os.path.dirname(os.path.abspath(__file__)))
    runner = ScrapyRunner()
    runner.run_all_spiders(runner.get_all_spider_names())


import smtplib
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.header import Header
from email.utils import formataddr
from typing import Sequence, Union

MsgType = Union[MIMEText, MIMEMultipart]

class Email:
    def __init__(self, host: str, port: int, username: str, authorization_code: str):
        self.host = host
        self.port = port  # 465
        self.username = username
        self.authorization_code = authorization_code  # Not email password
        self.smtp_obj = smtplib.SMTP_SSL(self.host, self.port)
        self.login()

    def login(self) -> None:
        self.smtp_obj.login(self.username, self.authorization_code)

    @staticmethod
    def set_headers(
        msg: MsgType,
        sender_name: str,
        sender_email: str,
        receiver_name: str,
        receiver_email: str,
        subject: str,
    ) -> MsgType:
        msg["From"] = formataddr((str(Header(sender_name, "utf-8")), sender_email))
        msg["To"] = formataddr((str(Header(receiver_name, "utf-8")), receiver_email))
        msg["Subject"] = Header(subject, "utf-8")
        return msg

    def send(self, msg: Union[MsgType, str], to_addrs: Union[str, Sequence[str]]) -> None:
        raw_msg = msg.as_string() if not isinstance(msg, str) else msg
        self.smtp_obj.sendmail(self.username, to_addrs, raw_msg)


_registry = {}

def register_request_class(cls):
    name = cls.__name__
    if name in _registry:
        raise ValueError(f"Request class '{name}' already registered.")
    _registry[name] = cls
    return cls

def get_request_class(class_name: str):
    cls = _registry.get(class_name)
    if cls is None:
        raise ValueError(f"Request class '{class_name}' not found in registry.")
    return cls

def is_registered(class_name: str) -> bool:
    return class_name in _registry

def clear_registry():
    _registry.clear()

def all_registered_names() -> list:
    return list(_registry.keys())



# pytypes.py
from typing import Any, Dict, List, TypedDict, Union
# We say messages can have any value
# Functions we define may have fixed types, but someone could add a type
# function that outputs any arbitrary object
Message = Dict[Union[str, int], Any]

TypeDefDict = Dict[str, "FieldDefDict"]

FieldDefDict = TypedDict(
    "FieldDefDict",
    {
        "name": str,
        "type": str,
        "message_type_name": str,
        "message_typedef": TypeDefDict,
        "alt_typedefs": Dict[str, str | TypeDefDict],
        "example_value_ignored": Any,
        "seen_repeated": bool,
        "field_order": List[str],
    },
    total=False,
)


import math
from typing import Union

def fnv1a_hash(data: Union[str, bytes], seed: int = 0) -> int:
    if isinstance(data, str):
        data = data.encode("utf-8")
    hash_ = 2166136261 ^ seed
    for b in data:
        hash_ ^= b
        hash_ = (hash_ * 16777619) & 0xFFFFFFFF
    return hash_

def set_bit(array: bytearray, index: int):
    array[index // 8] |= 1 << (index % 8)

def get_bit(array: bytearray, index: int) -> bool:
    return (array[index // 8] >> (index % 8)) & 1

class BloomFilterPy:
    def __init__(self, size: int, expected: int, hash_count: int = 0):
        self.size = size
        self.bytes = (size + 7) // 8
        self.bit_array = bytearray(self.bytes)
        self.inserted = 0
        if hash_count == 0:
            self.hash_count = max(1, round(size / expected * math.log(2)))
        else:
            self.hash_count = hash_count

    def add(self, data: Union[str, bytes]):
        for i in range(self.hash_count):
            h = fnv1a_hash(data, i)
            index = h % self.size
            set_bit(self.bit_array, index)
        self.inserted += 1

    def exists(self, data: Union[str, bytes]) -> bool:
        for i in range(self.hash_count):
            h = fnv1a_hash(data, i)
            index = h % self.size
            if not get_bit(self.bit_array, index):
                return False
        return True
    
    def free(self):
        """
        C extension previously used for ~30% faster inserts,
        but due to strong OS and Python version dependencies,
        related compiled files need to be built manually.
        This interface is the fallback version for the C extension,
        keeping full API compatibility.
        """
        pass

    def clear(self):
        self.bit_array = bytearray(self.bytes)
        self.inserted = 0

    def estimated_false_positive_rate(self) -> float:
        if self.inserted == 0:
            return 0.0
        m = self.size
        k = self.hash_count
        n = self.inserted
        return (1.0 - math.exp(-k * n / m)) ** k

    def get_hash_count(self) -> int:
        return self.hash_count

    def get_indices(self, data: Union[str, bytes]) -> list[int]:
        indices = []
        for i in range(self.hash_count):
            h = fnv1a_hash(data, i)
            indices.append(h % self.size)
        return indices
    
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.free()

try:
    from wrapper import BloomFilter as _C_BloomFilter
    BloomFilter = _C_BloomFilter
except (ImportError, OSError):
    # fallback
    BloomFilter = BloomFilterPy

__all__ = [
    "BloomFilter"
]

if __name__ == "__main__":
    bf = BloomFilter(size=1000, expected=100, hash_count=0)
    print(bf.get_hash_count())
    bf.add("asgshyjhyt")
    bf.add("asgyiylo98")
    bf.add("asgyiylo96687")
    print(bf.get_indices("daafasg"))


import asyncio, time
from ..extensions import signals, SignalInfo
from .downloader import *
from ..exceptions import DownloadError
from ..interceptors import ChainResult, ChainNextEnum
from ..interceptors.chains import _ensure_asyncgen
from ..utils.concurrency import safe_call
from typing import TYPE_CHECKING, Dict, Union
if TYPE_CHECKING:
    from ..crawler import Crawler
    from .tasks import TaskManager
    from .scheduler import Scheduler
    from ..extensions import SignalManager
    from ..settings import SettingsInfo
    from ..item import Item
    from ..interceptors import ChainManager, InterruptibleChainManager
    from ..spiders import Spider
    from .sessions import SessionManager, SessionWrapper, WebSocketEntry, CloseSignal

class Engine:
    def __init__(self, crawler: "Crawler", spider: "Spider"):
        self.stop_event: asyncio.Event = crawler.stop_event
        self.taskManager: "TaskManager" = crawler.taskManager
        self.settings: "SettingsInfo" = crawler.settings

        self.sessions: "SessionManager" = crawler.sessions
        self.sessions_lock: asyncio.Lock = crawler.sessions_lock

        self.signalManager: "SignalManager" = crawler.signalManager
        self.scheduler: "Scheduler" = crawler.scheduler
        self.downloader: "Downloader" = crawler.downloader
        self.spider: "Spider" = spider
        self.spiderInterceptor_chain: "InterruptibleChainManager" = crawler.spiderInterceptor_chain
        self.downloadInterceptor_chain: "InterruptibleChainManager" = crawler.downloadInterceptor_chain
        self.pipelines_chain: "ChainManager" = crawler.pipelines_chain

        from ..utils import init_logger
        self.logger = init_logger(log_info=self.settings.LOG_INFO, logger_name=__name__)
        if crawler.kafkaManager:
            from ..utils import KafkaLoggingHandler
            kafka_handler = KafkaLoggingHandler(kafka=crawler.kafkaManager, stop_event=self.stop_event).create_fmt(self.settings)
            self.logger.addHandler(kafka_handler)

    @classmethod
    def from_crawler(cls, crawler: "Crawler", spider: "Spider"):
        return cls(
            crawler=crawler,
            spider=spider,
        )

    async def start(self, *args, **kwargs):
        self.signalManager.send(signal=signals.engine_started, data=SignalInfo(signal_time=time.time()))
        if self.pipelines_chain.chain_list: # In fact, at least one pipeline component must be registered, otherwise there may be bugs
            await self.pipelines_chain.forward_pass(call_func_cls=self.pipelines_chain.chain_list[0].instance, call_func_name="open_spider", pad_data=self.spider)
        self.signalManager.send(signal=signals.spider_opened, data=SignalInfo(spider=self.spider, signal_time=time.time()))

        # Retrieve requests directly from the spider's start method without additional processing,
        # mark them as start URLs, and submit them to the spider middleware chain.
        async for output in self.spider.start(*args, **kwargs):
            await self.taskManager.create(coro=self.get_spider_output(output=output, mark_as_start=True))
        try:
            await self.taskManager.wait_until_stopped()
        except KeyboardInterrupt as e:
            async with self.taskManager.lock:
                self.taskManager.active_tasks -= 1

        if self.pipelines_chain.chain_list:
            await self.pipelines_chain.forward_pass(call_func_cls=self.pipelines_chain.chain_list[0].instance, call_func_name="close_spider", pad_data=self.spider)
        await self.signalManager._safe_put(signal=signals.spider_closed, data=SignalInfo(spider=self.spider, signal_time=time.time()))
        await self.signalManager._safe_put(signal=signals.engine_stopped, data=SignalInfo(signal_time=time.time()))

    async def get_spider_output(self, output, response=None, mark_as_start=False):
        async for single_result in _ensure_asyncgen(output):
            if isinstance(single_result, Request) and mark_as_start:
                single_result.meta["is_start_url"] = True
            async for item in self.spiderInterceptor_chain.process_spider_output_chain(
                response=response,
                result=single_result,
                spider=self.spider
            ):
                await self.taskManager.create(coro=self.manager_spiderinterceptors_result(spiderinterceptors_result=item))

    async def manager_spiderinterceptors_result(self, spiderinterceptors_result: ChainResult):
        if spiderinterceptors_result.next == ChainNextEnum.RESCHEDULE:
            await self.taskManager.create(coro=self.process_scheduler(request=spiderinterceptors_result.request))
        elif spiderinterceptors_result.next == ChainNextEnum.SPIDER:
            await self.process_response(response=spiderinterceptors_result.response, request=spiderinterceptors_result.request)
        elif spiderinterceptors_result.next == ChainNextEnum.PIPELINE:
            await self.process_items(item=spiderinterceptors_result.item)
        elif spiderinterceptors_result.next == ChainNextEnum.EXCEPTION:
            if not spiderinterceptors_result.is_across:
                await self.spiderInterceptor_chain.process_spider_exception_chain(
                    response=spiderinterceptors_result.response, 
                    exception=spiderinterceptors_result.exception, 
                    spider=spiderinterceptors_result.spider, 
                    callback=self.manager_spiderinterceptors_result,
                    is_across=1
                )
            else:
                data = spiderinterceptors_result.model_dump().copy()
                data["signal_time"] = time.time()
                self.signalManager.send(signal=signals.spider_error, data=SignalInfo(**data))
        elif spiderinterceptors_result.next == ChainNextEnum.SESSION:
            if spiderinterceptors_result.signal.websocket_end_for_key or spiderinterceptors_result.signal.websocket_end_for_url:
                self.end_websocket(signal=spiderinterceptors_result.signal)
            elif spiderinterceptors_result.signal.session_end:
                self.sessions.mark_end(session_id=spiderinterceptors_result.signal.session_id)

    def end_websocket(self, signal: "CloseSignal"):
        wrapper: "SessionWrapper" = self.sessions.get_or_create_session(signal.session_id)
        if signal.websocket_end_for_url:
            websocket_entry: "WebSocketEntry" = wrapper.websocket_pool.get_from_url(signal.websocket_end_for_url)
            if not websocket_entry:
                return
            end_url = websocket_entry.url
        elif signal.websocket_end_for_key: # In fact, the spiderInterceptor already has a key ->URL
            websocket_entry: "WebSocketEntry" = wrapper.websocket_pool.get_from_key(signal.websocket_end_for_key)
            if not websocket_entry:
                return
            end_url = websocket_entry.url
        else:
            return
        wrapper.websocket_pool.mark_end_from_url(end_url)
        websocket_entry.release()
        self.sessions.release(signal.session_id)

    # Scheduler: repeatedly fetches and processes requests until none are left.
    async def process_scheduler(self, request: Request=None):
        if self.stop_event.is_set():
            return
            
        if request:
            put_scheduler = await self.scheduler.put(request=request, spider=self.spider)
            if put_scheduler:
                await self.taskManager.create(coro=self.process_scheduler())
            else:
                return

        if self.scheduler.is_distributed:
            request = await self.scheduler.get(spider=self.spider)
            if isinstance(request, int) and not request: # scheduler empty
                self.signalManager.send(signal=signals.scheduler_empty, data=SignalInfo(signal_time=time.time()))
                return
            elif isinstance(request, Request):
                await self.taskManager.create(coro=self.process_downloadInterceptor_chain(request=request))
            elif isinstance(request, int):
                await self.taskManager.create(coro=self.process_scheduler())
        else:
            try:
                request = await asyncio.wait_for(self.scheduler.get(spider=self.spider), timeout=1.0)
                await self.taskManager.create(coro=self.process_downloadInterceptor_chain(request=request))
            except asyncio.TimeoutError:
                pass
            except asyncio.CancelledError:
                raise
            if self.scheduler.empty(spider=self.spider):
                self.signalManager.send(signal=signals.scheduler_empty, data=SignalInfo(signal_time=time.time()))
                return 
            else:
                await self.taskManager.create(coro=self.process_scheduler())

    # Download middleware processing
    async def process_downloadInterceptor_chain(self, response: Union[Response, BaseException, None]=None, request: Request=None):
        if response:
            await self.downloadInterceptor_chain.response_intercept_chain(request=request, response=response, spider=self.spider, callback=self.manager_downloadinterceptors_result)
        elif request:
            await self.downloadInterceptor_chain.request_intercept_chain(request=request, spider=self.spider, callback=self.manager_downloadinterceptors_result)
    
    # Handles results from the download interceptors or exceptions raised during downloading.
    async def manager_downloadinterceptors_result(self, downloadinterceptors_result: ChainResult):
        if downloadinterceptors_result.next == ChainNextEnum.RESCHEDULE:
            await self.taskManager.create(coro=self.process_scheduler(request=downloadinterceptors_result.request))
        elif downloadinterceptors_result.next == ChainNextEnum.DOWNLOADER:
            await self.taskManager.create(coro=self.process_downloader(request=downloadinterceptors_result.request))
        elif downloadinterceptors_result.next == ChainNextEnum.RESPONSE:
            await self.taskManager.create(coro=self.process_downloadInterceptor_chain(response=downloadinterceptors_result.response, request=downloadinterceptors_result.request))
        elif downloadinterceptors_result.next == ChainNextEnum.SPIDER:
            await self.spiderInterceptor_chain.process_spider_input_chain(
                response=downloadinterceptors_result.response, 
                request=downloadinterceptors_result.request, 
                spider=self.spider, 
                callback=self.manager_spiderinterceptors_result
            )
        elif downloadinterceptors_result.next == ChainNextEnum.EXCEPTION:
            if not downloadinterceptors_result.is_across:
                await self.downloadInterceptor_chain.exception_intercept_chain(request=downloadinterceptors_result.request, exception=downloadinterceptors_result.exception, spider=downloadinterceptors_result.spider, callback=self.manager_downloadinterceptors_result, is_across=1)
            else:
                await self.process_response(response=downloadinterceptors_result.exception, request=downloadinterceptors_result.request)

    # Downloader
    async def process_downloader(self, request: Request):
        if isinstance(request, HttpRequest):
            await self.downloader.fetch_http(request=request, callback=self.process_downloadInterceptor_chain)
        elif isinstance(request, WebSocketRequest):
            await self.taskManager.create(coro=self.process_websocket_request(request=request))
            
    # Callback to spider with the response
    async def process_response(self, response: Union[Response, BaseException], request: Request):
        if isinstance(response, BaseException):
            output = self.get_backFunc(backFunc=request.errback, response=response, fill_text=f"Response error {str(response)} with no errback provided, ignoring this request")
            if not output:
                return
        elif isinstance(response, Response):
            await self.scheduler.put_is_req(request=request, spider=self.spider)
            output = self.get_backFunc(backFunc=request.callback, response=response)
        else:
            return
        if not self.stop_event.is_set():
            await self.taskManager.create(coro=self.get_spider_output(output=output, response=response))

    # manager callback
    def get_backFunc(self, backFunc=None, response: Union[Response, BaseException]=None, fill_text=""):
        if isinstance(backFunc, str):
            callbackFunc = getattr(self.spider, backFunc, None)
        elif callable(backFunc):
            callbackFunc = backFunc
        else:
            callbackFunc = None
            
        if callbackFunc:
            output = callbackFunc(response) if response else callbackFunc()
            return output
        else:
            if isinstance(response, HttpResponse):
                self.logger.info(f"Response succeeded with text: {response.text}, no callback provided, task finished.")
            elif isinstance(response, WebSocketResponse):
                self.logger.info(f"Response succeeded with message: {response.msg[0]}, no callback provided, task finished.")
            else:
                self.logger.info(fill_text)

    async def process_items(self, item: Union["Item", Dict]):
        self.signalManager.send(signal=signals.item_scraped, data=SignalInfo(signal_time=time.time(), item=item, spider=self.spider))
        await self.pipelines_chain.forward_pass(call_func_cls=self.pipelines_chain.chain_list[0].instance, call_func_name="process_item", pad_data=item, spider=self.spider)

    # Process a WebSocket request
    async def process_websocket_request(self, request: WebSocketRequest):
        wrapper: "SessionWrapper" = self.sessions.get_or_create_session(request.session_id, cookies=request.cookies)
        if not request.url:
            raise ValueError("Scheduling logic error: this request was not properly processed by spider middleware")
        
        websocket_entry: "WebSocketEntry" = wrapper.get_websocket(request.url)
        if websocket_entry and request.send_message:
            if getattr(getattr(websocket_entry.websocket, "curl", None), "_curl", None) is not None:
                # WebSocket communication is deduplicated by connection only.
                # Once connection uniqueness is ensured, subsequent messages run on a single device,
                # so message deduplication is handled only by new_req_seen, not in is_req.
                # Send the message asynchronously in a thread to avoid blocking.

                # await run_with_timeout(ws.send, request.send_message, stop_event=self.stop_event, timeout=3)
                for msg in request.send_message:
                    await safe_call(websocket_entry.websocket.send, msg.data, flags=msg.flags)

            websocket_entry.release()
            self.sessions.release(request.session_id)
            return
        await self.do_websocket_connect(wrapper=wrapper, connect_request=request)

    async def do_websocket_connect(self, wrapper: "SessionWrapper", connect_request: WebSocketRequest):
        try:
            task, queue, websocket_event = await self.downloader.fetch_websocket(wrapper, connect_request)
            while (not self.stop_event.is_set()) and (not websocket_event.is_set()):
                try:
                    msg = await queue.get()
                    # msg = await asyncio.wait_for(queue.get(), timeout=3.0)
                    if isinstance(msg, DownloadError) or (isinstance(msg, str) and msg == self.settings.WS_END_TAG):
                        websocket_event.set()
                        # task.cancel()
                        # self.logger.debug(f'{msg}: {connect_request.url}, listener ended')
                        break
                    await self.taskManager.create(coro=self.process_downloadInterceptor_chain(response=msg, request=connect_request))
                    if b'keepalive ping timeout' in msg.msg[0]:
                        websocket_event.set()
                    await asyncio.sleep(0)
                except asyncio.TimeoutError:
                    pass
        except asyncio.CancelledError:
            self.logger.debug(f'WebSocket listener task cancelled: {connect_request.url}')
            raise
        except BaseException as e:
            results = DownloadError(exception=e, request=connect_request)
            await self.taskManager.create(coro=self.process_downloadInterceptor_chain(response=results, request=connect_request))
            self.end_websocket(signal=CloseSignal(session_id=connect_request.session_id, websocket_end_for_url=connect_request.url))
        finally:
            # Release session regardless of normal exit or exception cancellation
            self.sessions.release(connect_request.session_id) # Release the session
            await wrapper.close_websocket(connect_request.url) # Close underlying connection


# api.py
from . import length_delim
from .schema import TypeDef
from .config import default as default_config
from .exceptions import (
    DecoderException,
    EncoderException
)
from typing import Union, Tuple, Optional, TYPE_CHECKING

# Circular imports on Config if we don't check here
if TYPE_CHECKING:
    from .pytypes import Message, TypeDefDict
    from .config import Config

def decode_message(
    buf: bytes, 
    message_type: Optional[Union[str, "TypeDefDict"]]=None, 
    config: Optional["Config"]=None
) -> Tuple["Message", "TypeDefDict"]:
    """Decode a protobuf message and return a python dictionary representing
    the message.

    Args:
        buf: Bytes representing an encoded protobuf message
        message_type: Optional type to use as the base for decoding. Allows for
            customizing field types or names. Can be a python dictionary or a
            message type name which maps to the `known_types` dictionary in the
            config. Defaults to an empty definition '{}'.
        config: `.config.Config` object which allows
            customizing default types for wire types and holds the
            `known_types` array. Defaults to
            `.config.default` if not provided.
    Returns:
        A tuple containing a python dictionary representing the message and a
        type definition for re-encoding the message.

        The type definition is based on the `message_type` argument if one was
        provided, but may add additional fields if new fields were encountered
        during decoding.
    """

    if config is None:
        config = default_config

    if isinstance(buf, bytearray):
        buf = bytes(buf)
    elif isinstance(buf, str):
        buf = buf.encode("utf-8")
    
    if not isinstance(buf, bytes):
        raise TypeError("not expecting type '%s'" % type(buf))

    if message_type is None:
        message_type = {}
    elif isinstance(message_type, str):
        message_type = config.known_types.get(message_type, {})

    if not isinstance(message_type, dict):
        raise DecoderException("Decode message received an invalid typedef type. Typedef should be a string with a message name, a dictionary, or None")
    
    value, typedef, _, _ = length_delim.decode_message(
        buf, config, TypeDef.from_dict(message_type)
    )
    return value, typedef.to_dict()


def encode_message(
    value: "Message", 
    message_type: Union[str, "TypeDefDict"], 
    config: Optional["Config"]=None
) -> bytes:
    """Re-encode a python dictionary as a binary protobuf message.

    Args:
        value: Python dictionary to re-encode to bytes. This should usually be
            a modified version of the dictionary returned by `decode_message`.
        message_type: Type definition to use to re-encode the message. This
            will should generally be the type definition returned from the
            original `decode_message` call.
        config: `.config.Config` object which allows
            customizing default types for wire types and holds the
            `known_types` array. Defaults to
            `.config.default` if not provided.
    Returns:
        A bytearray containing the encoded protobuf message.
    """

    if config is None:
        config = default_config

    if message_type is None:
        raise EncoderException(
            "Encode message must have valid type definition. message_type cannot be None"
        )

    if isinstance(message_type, str):
        if message_type not in config.known_types:
            raise EncoderException(
                "The provided message type name (%s) is not known. Encoding requires a valid type definition"
                % message_type
            )
        message_type = config.known_types[message_type]

    if not isinstance(message_type, dict):
        raise EncoderException(
            "Encode message received an invalid typedef type. Typedef should be a string with a message name or a dictionary."
        )
    return bytes(
        length_delim.encode_message(
            value, config, TypeDef.from_dict(message_type)
        )
    )


from typing import Protocol, Callable, Any, TypeVar, Union, Awaitable

T = TypeVar("T")

class SignalHooks(Protocol):
    SignalCallback = Union[Callable[[T], Any], Callable[[T], Awaitable[Any]]]
    def connect(self, signal: object, callback: SignalCallback) -> None: ...

class SignalsHooks(Protocol):
    signals: SignalHooks



# config.py
from .schema import WIRE_TYPE_DEFAULTS
from .exceptions import DecoderException
from typing import TYPE_CHECKING, Dict
if TYPE_CHECKING:
    from .pytypes import TypeDefDict

class Config:
    def __init__(self):
        # Map of message type names to typedefs, previously stored at
        # `blackboxprotobuf.known_messages`
        self.known_types: Dict[str, "TypeDefDict"] = {}

        # Default type for "bytes" like objects that aren't messages or strings
        # Other option is currently just 'bytes_hex'
        self.default_binary_type = "bytes"

        # Change the default type for a wiretype (eg. change ints to be signed
        # by default or fixed fields to always be float)
        self.default_types: Dict[int, str] = {}

        # Configure whether bbpb should try to re-encode fields in the same
        # order they decoded
        # Field order shouldn't matter for real protobufs, but is there to ensure
        # that bytes/string are accidentally valid protobufs don't get scrambled
        # by decoding/re-encoding
        self.preserve_field_order = True

    def get_default_type(self, wiretype: int) -> str:
        default_type = self.default_types.get(wiretype, None)
        if default_type is None:
            default_type = WIRE_TYPE_DEFAULTS.get(wiretype, None)

        if default_type is None:
            raise DecoderException(
                "Could not find default type for wire type %d" % wiretype
            )
        return default_type


default = Config()



# length_delim.py
import logging
from .schema import BaseWireTypes, WireTypes, ENCODERS, DECODERS, TypeDef, FieldDef
from .utils import wire
from .exceptions import (
    EncoderException,
    DecoderException,
    TypedefException,
    BlackboxProtobufException,
)

from typing import Any, Callable, Dict, Tuple, Optional, List, TYPE_CHECKING, Union, Any
if TYPE_CHECKING:
    from .config import Config
    from .pytypes import Message

logger = logging.getLogger(__name__)

def encode_message(
    data: "Message", 
    config: "Config", 
    typedef: TypeDef, 
    path: Optional[List[str]]=None, 
    field_order: Optional[List[str]]=None
) -> bytes:
    """Encode a Python dictionary to a binary protobuf message"""
    output = bytearray()
    if path is None:
        path = []

    output_len = 0
    field_outputs: Dict[str, List[bytes]] = {}
    for field_id, value in data.items():
        field_number, outputs = _encode_message_field(
            config, typedef, path, field_id, value
        )

        # In case the field number is represented in multiple locations in data
        # (eg. as an int, as name, as a string with an int)
        field_outputs.setdefault(field_number, []).extend(outputs)
        output_len += len(outputs)

    if output_len > 0:
        if (
            config.preserve_field_order
            and field_order is not None
            and len(field_order) == output_len
        ):
            # check for old typedefs which had field_order as a tuple
            if isinstance(field_order[0], tuple):
                field_order = [x[0] for x in field_order]
            for field_number in field_order:
                try:
                    output += field_outputs[field_number].pop(0)
                except (IndexError, KeyError):
                    # If these don't match up despite us checking the overall
                    # length, then we probably have something weird going on
                    # with field naming.
                    # This might mean ordering is off from the original, but
                    # should break real protobuf messages
                    logger.debug(
                        "The field_order list does not match the fields from _encode_message_field"
                    )
                    # If we're hitting a mismatch between the field order and
                    # what data we have, then just bail. We can encode the rest
                    # normally
                    break

        # Group  together elements in an array
        for values in field_outputs.values():
            for value in values:
                output += value

    return output

def _encode_message_field(
    config: "Config", 
    typedef: TypeDef, 
    path: List[str], 
    field_id: Union[str, int], 
    value: Any
) -> Tuple[str, List[bytes]]:
    field_key = str(field_id)

    fielddef_results = typedef.lookup_fielddef(field_key)

    if fielddef_results is None:
        raise EncoderException(
            "Provided field name/number %s is not valid" % (field_key),
            path,
        )
    field_number, fielddef = fielddef_results

    field_path = path[:]
    field_path.append(str(field_number))

    field_type = fielddef.lookup_field_type(field_key, config, field_path)

    if field_type is None:
        raise EncoderException(
            "Provided field name/number %s / %s is not valid"
            % (field_key, field_number),
            field_path,
        )

    field_encoder: Optional[Callable[[Any], bytes]] = None
    if isinstance(field_type, TypeDef):
        field_typedef = field_type
        field_type = "message"
        field_encoder = lambda data: encode_lendelim_message(
            data,
            config,
            field_typedef,
            path=field_path,
            field_order=fielddef.field_order,
        )
    else:
        if field_type not in ENCODERS:
            raise TypedefException("Unknown type: %s" % field_type)
        field_encoder = ENCODERS[field_type]
        if field_encoder is None:
            raise TypedefException(
                "Encoder not implemented for %s" % field_type, field_path
            )

    # Encode the tag
    tag = wire.encode_tag(
        int(field_number), WireTypes[field_type].value
    )

    outputs = []
    try:
        # Repeated values we'll encode each one separately and add them to the outputs list
        # Packed values take in a list, but encode them into a single length
        # delimited field, so we handle those as a non-repeated value
        if isinstance(value, list) and not field_type.startswith("packed_"):
            for repeated in value:
                outputs.append(tag + field_encoder(repeated))
        else:
            outputs.append(tag + field_encoder(value))
            
    except EncoderException as exc:
        exc.set_path(field_path)
        raise exc.with_traceback(exc.__traceback__)

    return field_number, outputs

def decode_message(
    buf: bytes, 
    config: "Config", 
    typedef: Optional[TypeDef]=None, 
    pos: int=0, 
    end: Optional[int]=None, 
    depth: int=0, 
    path: Optional[List[str]]=None
) -> Tuple["Message", TypeDef, List[str], int]:
    """Decode a protobuf message with no length prefix"""
    if end is None:
        end = len(buf)

    if typedef is None:
        typedef = TypeDef()

    if path is None:
        path = []

    output: "Message" = {}
    seen_repeated: Dict[str, bool] = {}
    mut_typedef = typedef.make_mutable()

    grouped_fields, field_order, pos = _group_by_number(buf, pos, end, path)
    for field_number, (wire_type, buffers) in grouped_fields.items():
        # wire_type should already be validated by _group_by_number

        field_path = path[:] + [field_number]

        fielddef_pair: Optional[Tuple[str, FieldDef]] = mut_typedef.lookup_fielddef_number(field_number)

        if fielddef_pair is None:
            fielddef = FieldDef(field_number)
        else:
            fielddef = fielddef_pair[1]

        # Decode messages (which may have multiple typedefs)  or unknown length delimited fields
        if wire_type == BaseWireTypes.LENGTH_DELIMITED and not isinstance(
            fielddef.lookup_field_type_number("0", config, field_path), str
        ):
            output_map, new_fielddef = _try_decode_lendelim_fields(
                buffers, fielddef, config, field_path
            )

            # Merge length delim field into the output map
            for field_key, field_outputs in output_map.items():
                output.setdefault(field_key, []).extend(field_outputs)
            seen_repeated[fielddef.name] = new_fielddef.seen_repeated
            mut_typedef.set_fielddef(field_number, new_fielddef)
        else:
            field_outputs, new_fielddef, field_alt_type_id = _decode_standard_field(
                wire_type, buffers, fielddef, config, path
            )

            field_key = new_fielddef.field_key(field_alt_type_id)
            output.setdefault(field_key, []).extend(field_outputs)
            seen_repeated[fielddef.name] = new_fielddef.seen_repeated

            # Save the field typedef/type back to the typedef
            mut_typedef.set_fielddef(field_number, new_fielddef)

    _simplify_output(output, seen_repeated)
    return output, mut_typedef, field_order, pos

def _decode_standard_field(
    wire_type: int, 
    buffers: List[bytes], 
    fielddef: FieldDef, 
    config: "Config", 
    field_path: List[str]
) -> Tuple[List[Any], FieldDef, str]:
    field_outputs = None
    field_alt_type_id = None
    for alt_type_id, field_type in fielddef.resolve_types(config, field_path).items():
        if isinstance(field_type, TypeDef):
            # Skip message types
            continue
        if (
            not isinstance(field_type, str)
            or BaseWireTypes[field_type] != wire_type
        ):
            raise DecoderException(
                "Type %s from typedef did not match wiretype %s"
                % (field_type, wire_type),
                path=field_path,
            )

        if field_type not in DECODERS:
            raise TypedefException(
                "Type %s does not have a decoder" % (field_type),
                path=field_path,
            )
        decoder = DECODERS[field_type]
        try:
            field_outputs = [decoder(buf, 0)[0] for buf in buffers]
            field_alt_type_id = alt_type_id
        except BlackboxProtobufException as exc:
            # Error decoding, try next one if we have one
            continue
        # Decoding worked
        break

    if field_outputs is None:
        field_type = config.get_default_type(wire_type)
        default_decoder = DECODERS[field_type]

        field_outputs = [default_decoder(buf, 0)[0] for buf in buffers]

    mut_fielddef = fielddef.make_mutable()
    if field_alt_type_id is None:
        field_alt_type_id = mut_fielddef.next_alt_type_id()

    mut_fielddef.set_type(field_alt_type_id, field_type)

    if field_outputs is None:
        raise DecoderException(
            "Unable to decode wire_type %s" % (wire_type),
            path=field_path,
        )
    if isinstance(field_type, str) and field_type.startswith("packed_"):
        # Packed decoding will return a list of lists
        field_outputs = [y for x in field_outputs for y in x]
        mut_fielddef.mark_repeated()
    # Mark repeated if we have have more than one
    # Don't need to worry if it's already repeated
    elif len(field_outputs) > 1:
        mut_fielddef.mark_repeated()

    return field_outputs, mut_fielddef, field_alt_type_id

def _simplify_output(output: "Message", seen_repeated: Dict[str, bool]) -> None:
    # If any outputs only have one element, convert them from a list to solo
    # Mutates output
    for field_key, field_outputs in output.items():
        if isinstance(field_outputs, list) and len(field_outputs) == 1:
            field_name = (
                field_key.split("-", 1)[0]
                if isinstance(field_key, str)
                else str(field_key)
            )
            if not seen_repeated[field_name]:
                output[field_key] = field_outputs[0]

def _group_by_number(
    buf: bytes, 
    pos: int, 
    end: int, 
    path: List[str]
) -> Tuple[Dict[str, Tuple[int, List[bytes]]], List[str], int]:
    # Parse through the whole message and split into buffers based on wire
    # type and organized by field number. This forces us to parse the whole
    # message at once, but I think we're doing that anyway. This catches size
    # errors early as well, which is usually the best indicator of if it's a
    # protobuf message or not.
    # Returns a dictionary like:
    #     {
    #         "2": (<wiretype>, [<data>])
    #     }

    output_map: Dict[str, Tuple[int, List[bytes]]] = {}
    field_order = []
    while pos < end:
        # Read in a field
        field_number, wire_type, pos = wire.decode_tag(buf, pos)

        # We want field numbers as strings everywhere
        field_id = str(field_number)

        field_path = path[:] + [field_id]

        if field_id in output_map and output_map[field_id][0] != wire_type:
            # This should never happen
            raise DecoderException(
                "Field %s has mistmatched wiretypes. Previous: %s Now: %s"
                % (field_id, output_map[field_id][0], wire_type),
                path=field_path,
            )

        length = None
        if wire_type == BaseWireTypes.VARINT:
            # We actually have to read in the whole varint to figure out it's size
            _, new_pos = wire.decode_uvarint(buf, pos)
            length = new_pos - pos
        elif wire_type == BaseWireTypes.FIXED32:
            length = 4
        elif wire_type == BaseWireTypes.FIXED64:
            length = 8
        elif wire_type == BaseWireTypes.LENGTH_DELIMITED:
            # Read the length from the start of the message
            # add on the length of the length tag as well
            bytes_length, new_pos = wire.decode_varint(buf, pos)
            length = bytes_length + (new_pos - pos)
        elif wire_type in [
            BaseWireTypes.START_GROUP,
            BaseWireTypes.END_GROUP,
        ]:
            raise DecoderException("GROUP wire types not supported", path=field_path)
        else:
            raise DecoderException(
                "Got unknown wire type: %d" % wire_type, path=field_path
            )
        if pos + length > end:
            raise DecoderException(
                "Decoded length for field %s goes over end: %d > %d"
                % (field_id, pos + length, end),
                path=field_path,
            )

        field_buf = buf[pos : pos + length]

        if field_id in output_map:
            output_map[field_id][1].append(field_buf)
        else:
            output_map[field_id] = (wire_type, [field_buf])
        field_order.append(field_id)
        pos += length
    return output_map, field_order, pos

def _try_decode_lendelim_fields(
    buffers: List[bytes], 
    fielddef: FieldDef, 
    config: "Config", 
    path: List[str]
) -> Tuple["Message", FieldDef]:
    # Mutates message_output

    # This is where things get weird
    # To start, since we want to decode messages and not treat every
    # embedded message as bytes, we have to guess if it's a message or
    # not.
    # Unlike other types, we can't assume our message types are
    # consistent across the tree or even within the same message.
    # A field could be a bytes type that that decodes to multiple different
    # messages that don't have the same type definition. This is where
    # 'alt_typedefs' let us say that these are the different message types
    # we've seen for this one field.
    # In general, if something decodes as a message once, the rest should too
    # and we can enforce that across a single message, but not multiple
    # messages.
    # This is going to change the definition of "alt_typedefs" a bit from just
    # alternate message type definitions to also allowing downgrading to
    # 'bytes' or string with an 'alt_type' if it doesn't parse

    message_output: "Message" = {}

    # TODO potential performance improvement: Do a first pass with {} or just group by number and use the results to validate if it's even valid protobuf and quick match wire_types against typedefs
    try:
        outputs_map: Dict[str, Any] = {}
        field_order: List[str] = [] 

        next_alt_type_id = int(fielddef.next_alt_type_id())
        field_types = fielddef.resolve_types(config, path)

        # We don't want any mutable changes within this loop, we want
        # everything to rollback if it fails
        for buf in buffers:
            output = None
            output_typedef = None
            output_typedef_num = None
            new_field_order: List[str] = []

            for alt_type_id, field_type in sorted(
                field_types.items(), key=lambda x: int(x[0])
            ):
                # Skip non message types
                if not isinstance(field_type, TypeDef):
                    continue

                try:
                    (
                        output,
                        output_typedef,
                        new_field_order,
                        _,
                    ) = decode_lendelim_message(buf, config, field_type)
                except Exception as exc:
                    # If we get an exception, then this isn't the right typedef, try the next
                    continue

                output_typedef_num = alt_type_id
                # If we didn't get an exception, then we found the right type
                break
            # If we didn't find a type above, then try an anonymous type
            # If this fails, we fall back to string and bytes for all types
            if output is None:
                output, output_typedef, new_field_order, _ = decode_lendelim_message(
                    buf, config, None
                )
                output_typedef_num = str(next_alt_type_id)
                next_alt_type_id += 1

            if output_typedef is None or output_typedef_num is None:
                raise DecoderException(
                    "Could not find an output_typedef or output_typedef_num. This should not happen under any circumstances."
                )

            # save the output or typedef we found
            field_types[output_typedef_num] = output_typedef
            outputs_map.setdefault(output_typedef_num, []).append(output)

            # we should technically have a different field order for each instance of the data
            # but that would require a very messy JSON which we're trying to avoid
            if len(new_field_order) > len(field_order):
                field_order = new_field_order

        # was able to decode everything as a message
        mut_fielddef = fielddef.make_mutable()
        mut_fielddef.set_types(field_types)

        if config.preserve_field_order:
            mut_fielddef.set_field_order(field_order)

        # messages get set as "key-alt_number"
        for output_typedef_num, outputs in outputs_map.items():
            output_field_key = mut_fielddef.field_key(output_typedef_num)

            message_output[output_field_key] = outputs
            if len(outputs) > 1:
                mut_fielddef.mark_repeated()

        # success, return
        return message_output, mut_fielddef
    except DecoderException as exc:
        # this should be pretty common, don't be noisy or throw an exception
        logger.debug(
            "Could not decode a buffer for field (%s) as a message: %s",
            path,
            exc,
        )

    # Decoding as a message did not work, try strings and then the configured binary type
    # By default, default_binary_type will be redundant with bytes, but we want
    # to fall back on bytes if default_binary_type fails for any reason
    for target_type in ["string", config.default_binary_type, "bytes"]:
        try:
            outputs = []
            decoder = DECODERS[target_type]
            for buf in buffers:
                output, _ = decoder(buf, 0)
                outputs.append(output)

            field_alt_type_id = None
            # check if the type is already known
            field_types = fielddef.resolve_types(config, path)
            for alt_type_id, field_type in field_types.items():
                if field_type == target_type:
                    field_alt_type_id = alt_type_id
                    break

            mut_fielddef = fielddef.make_mutable()
            if field_alt_type_id is None:
                field_alt_type_id = mut_fielddef.add_type(target_type)

            field_key: str = mut_fielddef.field_key(field_alt_type_id)
            
            message_output[field_key] = outputs
            return message_output, mut_fielddef
        except DecoderException:
            continue

    # This should never happen, we should always be able to use bytes
    raise DecoderException("Unable to decode field with typedef", path=path)

def encode_lendelim_message(
    data: "Message", 
    config: "Config", 
    typedef: TypeDef, 
    path: Optional[List[str]]=None, 
    field_order: Optional[List[str]]=None
) -> bytes:
    """Encode data as a length delimited protobuf message"""
    message_out = encode_message(
        data, config, typedef, path=path, field_order=field_order
    )
    length = wire.encode_varint(len(message_out))
    return length + message_out


def decode_lendelim_message(
    buf: bytes, 
    config: "Config", 
    typedef: Optional[TypeDef]=None, 
    pos: int=0, 
    depth: int=0, 
    path: Optional[List[str]]=None
) -> Tuple["Message", TypeDef, List[str], int]:
    """Deocde a length delimited protobuf message from buf"""
    length, pos = wire.decode_varint(buf, pos)
    ret = decode_message(
        buf, config, typedef, pos, pos + length, depth=depth, path=path
    )
    return ret


from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ..hooks.signals import SignalsHooks

class Extension:
    def __init__(self, hooks: "SignalsHooks", **kwargs):
        self.hooks = hooks
        self.redisManager = kwargs.get("redisManager")
        self.mysqlManager = kwargs.get("mysqlManager")
        self.mongodbManager = kwargs.get("mongodbManager")
        self.rabbitmqManager = kwargs.get("rabbitmqManager")
        self.kafkaManager = kwargs.get("kafkaManager")

    @classmethod
    def from_crawler(cls, hooks: "SignalsHooks", **kwargs):
        return cls(hooks=hooks, **kwargs)


import json as jsonlib
import gzip, base64, warnings
from curl_cffi.const import CurlWsFlag
from urllib.parse import urlencode
from typing import Optional, Union, Dict, Tuple, List
from ....models.api import WebSocketMsg
from ....utils import ProtobufFactory
from .registry import register_request_class, get_request_class

class Request(object):
    ENCODE_FIELDS = set()

    def __init__(self,
        session_id="",
        url="",
        params=None,
        headers=None,
        cookies=None,
        proxies=None,
        timeout=30,
        allow_redirects=True,
        max_redirects=30,
        verify=None,
        impersonate=None,
        ja3=None,
        akamai=None,
        meta=None,
        dont_filter=None,
        callback=None,
        errback=None,
        desc_text="",
        no_proxy=False,
        **kwargs
    ) -> None:
        meta = meta or {}

        self.session_id = session_id
        self.url = self.join_url_params(url, params=params)
        self.headers = headers
        self.cookies = cookies
        self.proxies = proxies
        self.timeout=timeout
        self.allow_redirects=allow_redirects
        self.max_redirects=max_redirects
        self.verify=verify
        self.impersonate=impersonate
        self.ja3=ja3
        self.akamai=akamai
        self.meta = meta
        self.dont_filter = dont_filter
        self.callback = callback if isinstance(callback, str) else (callback.__name__ if callback else None)
        self.errback = errback if isinstance(errback, str) else (errback.__name__ if errback else None)
        self.desc_text = desc_text
        self.no_proxy = no_proxy
        self.kwargs = kwargs

    def is_protobuf(self):
        has_key = self.find_header_key(key="content-type")
        if not has_key:
            return False
        return any([it in self.headers[has_key].lower() for it in ["protobuf", "grpc"]])
    
    def find_header_key(self, key: str) -> Optional[str]:
        if self.headers:
            key_lower = key.lower()
            for k in self.headers:
                if k.lower() == key_lower:
                    return k
            return None

    def set_header(self, headers: Optional[Dict], key: str, value: str, mode: str = "append", sep: str = ", ") -> Dict:
        if headers is None:
            headers = {}
        real_key = self.find_header_key(key)
        if real_key is None:
            headers[key] = value
        else:
            if mode == "overwrite":
                headers[real_key] = value
            elif mode == "append":
                existing_value = headers[real_key]
                if value.lower() not in existing_value.lower():
                    headers[real_key] = f"{existing_value}{sep}{value}"
        return headers

    def join_url_params(self, url="", params=None):
        return f'{url}?{urlencode(params, doseq=True)}' if params else url
    
    def to_dict(self) -> Dict:
        def _encode_data(data):
            if isinstance(data, bytes):
                return {"__bytes__": True, "b64": base64.b64encode(data).decode()}
            elif isinstance(data, WebSocketMsg):
                return data.to_dict()
            elif isinstance(data, list):
                return [_encode_data(v) for v in data]
            elif isinstance(data, dict):
                return {k: _encode_data(v) for k, v in data.items()}
            elif isinstance(data, tuple):
                return tuple(_encode_data(v) for v in data)
            else:
                return data

        result = {}
        for key, value in self.__dict__.items():
            if key in self.ENCODE_FIELDS:
                result[key] = _encode_data(value)
            else:
                result[key] = value
        return result
    
    def to_bytes(self) -> bytes:
        d = self.to_dict()
        d["class"] = self.__class__.__name__
        json_bytes = jsonlib.dumps(d, ensure_ascii=False, separators=(",", ":")).encode("utf-8")
        return gzip.compress(json_bytes)

    @classmethod
    def from_bytes(cls, b: bytes) -> "Request":
        d = jsonlib.loads(gzip.decompress(b).decode("utf-8"))
        cls_name = d.pop("class", None)
        actual_cls: Union[HttpRequest, WebSocketRequest] = get_request_class(cls_name)
        return actual_cls._from_dict(d)
    
    @staticmethod
    def _decode_data(data):
        if isinstance(data, dict):
            if data.get("__bytes__"):
                return base64.b64decode(data["b64"])
            if data.get("__wsmsg__"):
                return WebSocketMsg.from_dict(data)
            return {k: Request._decode_data(v) for k, v in data.items()}
        elif isinstance(data, list):
            return [Request._decode_data(v) for v in data]
        elif isinstance(data, tuple):
            return tuple(Request._decode_data(v) for v in data)
        else:
            return data

    @classmethod
    def _from_dict(cls, d):
        kwargs = cls._decode_data(d.pop("kwargs", {}))
        for key in cls.ENCODE_FIELDS:
            if key in d:
                d[key] = cls._decode_data(d[key])
        return cls(**d, **kwargs)
    
@register_request_class
class HttpRequest(Request):
    ENCODE_FIELDS = {
        "headers", "data", "cookies", "proxies", "meta", "kwargs"
    }

    def __init__(self, 
        session_id="",
        url="", 
        params=None, 
        method="GET", 
        headers=None, 
        data=None, 
        json=None, 
        cookies=None, 
        proxies=None, 
        timeout=30,
        allow_redirects=True,
        max_redirects=30,
        verify=None,
        impersonate=None,
        ja3=None,
        akamai=None,
        meta=None, 
        dont_filter=None, 
        callback=None, 
        errback=None,
        desc_text="",
        no_proxy=False,
        **kwargs
    ):
        self.method = method.upper()
        if json:
            data = jsonlib.dumps(json, separators=(",", ":"))
            headers = self.set_header(headers=headers, key="Content-Type", value="application/json", mode="append", sep="; ")

        self.data = data
        super().__init__(
            session_id=session_id,
            url=url, 
            params=params, 
            headers=headers, 
            cookies=cookies, 
            proxies=proxies, 
            timeout=timeout,
            allow_redirects=allow_redirects,
            max_redirects=max_redirects,
            verify=verify,
            impersonate=impersonate,
            ja3=ja3,
            akamai=akamai,
            meta=meta, 
            dont_filter=dont_filter, 
            callback=callback, 
            errback=errback,
            desc_text=desc_text,
            no_proxy=no_proxy,
            **kwargs
        )
        if self.is_protobuf() and not isinstance(data, bytes):
            warnings.warn(f'[WARNING] Content-Type includes "protobuf", but the provided data is not in bytes format.')

    def protobuf_encode(self, typedef: Dict):
        self.data = ProtobufFactory.protobuf_encode(data=self.data, typedef=typedef)
        return self
    
    def grpc_encode(self, typedef_or_stream: Union[Dict, List[Tuple[Dict, Dict]]], is_gzip: bool=False):
        if isinstance(typedef_or_stream, dict):
            self.data = ProtobufFactory.grpc_encode(data=self.data, typedef=typedef_or_stream, is_gzip=is_gzip)
        elif isinstance(typedef_or_stream, list):
            self.data = ProtobufFactory.grpc_stream_encode(data=typedef_or_stream, is_gzip=is_gzip)
        return self
    
@register_request_class
class MediaRequest(HttpRequest):
    def __init__(self, 
        session_id="",
        url="", 
        params=None, 
        method="GET", 
        headers=None, 
        data=None, 
        json=None, 
        cookies=None, 
        proxies=None, 
        single_part_size=2999999, # The byte size of a segment
        media_size=0,
        timeout=30,
        allow_redirects=True,
        max_redirects=30,
        verify=None,
        impersonate=None,
        ja3=None,
        akamai=None,
        meta=None, 
        dont_filter=None, 
        callback=None, 
        errback=None,
        desc_text="",
        no_proxy=False,
        **kwargs
    ):
        super().__init__(
            session_id=session_id,
            url=url, 
            params=params, 
            method=method,
            headers=headers, 
            data=data,
            json=json,
            cookies=cookies, 
            proxies=proxies, 
            timeout=timeout,
            allow_redirects=allow_redirects,
            max_redirects=max_redirects,
            verify=verify,
            impersonate=impersonate,
            ja3=ja3,
            akamai=akamai,
            meta=meta, 
            dont_filter=dont_filter, 
            callback=callback, 
            errback=errback,
            desc_text=desc_text,
            no_proxy=no_proxy,
            **kwargs
        )
        self.single_part_size = single_part_size
        self.media_size = media_size
    
@register_request_class
class WebSocketRequest(Request):
    ENCODE_FIELDS = {
        "headers", "send_message", "ping_data", "cookies", "proxies", "meta", "kwargs"
    }

    def __init__(
        self,
        session_id="",
        websocket_id="",
        url="",
        params=None,
        headers=None,
        send_message: Union[WebSocketMsg, List[WebSocketMsg]] = None,
        ping_data: WebSocketMsg=None,
        ping_interval: float=15.0,
        cookies=None,
        proxies=None,
        timeout=30,
        allow_redirects=True,
        max_redirects=30,
        verify=None,
        impersonate=None,
        ja3=None,
        akamai=None,
        meta=None,
        dont_filter=None,
        callback=None,
        errback=None,
        desc_text="",
        no_proxy=False,
        **kwargs,
    ):
        super().__init__(
            session_id=session_id,
            url=url,
            params=params,
            headers=headers,
            cookies=cookies,
            proxies=proxies,
            timeout=timeout,
            allow_redirects=allow_redirects,
            max_redirects=max_redirects,
            verify=verify,
            impersonate=impersonate,
            ja3=ja3,
            akamai=akamai,
            meta=meta,
            dont_filter=dont_filter,
            callback=callback,
            errback=errback,
            desc_text=desc_text,
            no_proxy=no_proxy,
            **kwargs,
        )
        self.websocket_id = websocket_id
        if send_message is None:
            send_message = [WebSocketMsg(data=b"ping", flags=CurlWsFlag.BINARY)]
        elif not isinstance(send_message, list):
            send_message = [send_message]

        if not all([isinstance(it, WebSocketMsg) for it in send_message]):
            warning_text = f"send_message must be WebSocketMsg in versions >=0.2.2, got {type(send_message)}"
            warnings.warn(warning_text)
            raise ValueError(warning_text)
        
        if ping_data and not isinstance(ping_data, WebSocketMsg):
            raise ValueError(f"ping_data must be WebSocketMsg, got {type(ping_data)}")

        self.send_message = send_message

        self.ping_data = ping_data
        self.ping_interval = ping_interval

    def grpc_stream_encode(self) -> "WebSocketRequest":
        if not self.send_message:
            raise ValueError("No WebSocket messages to encode.")

        # all_flags = {m.flags for m in self.send_message}
        # if len(all_flags) > 1:
        #     raise ValueError(f"Cannot stream-encode messages with different flags: {all_flags}")

        for m in self.send_message:
            if not isinstance(m.data, (bytes, bytearray)):
                raise TypeError(f"Message {m!r} not encoded as bytes.")

        stream_bytes = b"".join(m.data for m in self.send_message)
        self.send_message = [WebSocketMsg(data=stream_bytes, flags=CurlWsFlag.BINARY)]
        return self


import asyncio
from scrapy_cffi.databases.mongodb import MongoDBManager

async def test():
    # Initialize MongoDBManager
    mongo = MongoDBManager(asyncio.Event(), "mongodb://localhost:27017", "test_db")
    await mongo.init()

    # Create a unique index on "name" if you want to prevent duplicates.
    # If not set, MongoDB allows duplicate entries, and insert_one may raise DuplicateKeyError if unique=True.
    await mongo.collection("test_collection").create_index("name", unique=True)

    # Insert documents
    await mongo.collection("test_collection").insert_one({"name": "Alice", "age": 23})
    await mongo.collection("test_collection").insert_one({"name": "Alice", "age": 23}) # May raise duplicate key error

    # Find one document
    doc = await mongo.collection("test_collection").find_one({"name": "Alice"})
    print(doc)
    print("————————————————————————————————————————")

    # Iterate over all documents in the collection
    tables_all = mongo.collection("test_collection").find()
    async for doc in tables_all:
        print(doc)

    # Drop the test database
    await mongo.drop_database("test_db")

    # Close the connection
    await mongo.close()

if __name__ == "__main__":
    asyncio.run(test())



from typing import Protocol, Dict, Any, TYPE_CHECKING, Union
if TYPE_CHECKING:
    from ..spiders import Spider

class SessionHooks(Protocol):
    def register_sessions(self, sessions: Dict[str, Any]) -> None: ...

    def get_session_cookies(self, session_id: str) -> Dict: ...

class SchedulerHooks(Protocol):
    def get_start_req(self, spider: "Spider", **kwargs) -> Union[None, bytes]: ...

class SpidersHooks(Protocol):
    session: SessionHooks
    scheduler: SchedulerHooks



from pydantic.dataclasses import dataclass
from ..core.downloader.internet import HttpRequest, WebSocketRequest, HttpResponse, WebSocketResponse
from ..spiders.base import BaseSpider
from ..item.base import Item
from typing import Union, Dict, Optional

@dataclass(config={"extra": "ignore", "arbitrary_types_allowed": True})
class SignalInfo:
    signal_time: Optional[float] = 0.0
    reason: Optional[str] = ""
    next: Optional[str] = ""
    response: Optional[Union[HttpResponse, WebSocketResponse]] = None
    exception: Optional[BaseException] = None
    spider: Optional[BaseSpider] = None
    request: Optional[Union[HttpRequest, WebSocketRequest]] = None
    item: Optional[Union[Item, Dict]] = None

__all__ = [
    "SignalInfo"
]


from pathlib import Path

def find_project_root(start: Path = None, is_demo=False) -> Path:
    if start is None:
        start = Path.cwd()
    all_path = [start / "demo"] if is_demo else [start, *start.parents]
    for path in all_path:
        if (path / "scrapy_cffi.toml").exists():
            return path
    raise FileNotFoundError("Project root not found (missing scrapy_cffi.toml)")


import hashlib, json, json5, orjson, re
from typing import Any, List, Dict, Union
import hashlib, json, json5, orjson, re, regex
from typing import Any, List, Dict, Union

def extract_nested_objects(
    text: str,
    key: str = "",
    re_rule: str = ""
) -> Union[List[Union[Dict, str]], Dict, str]:
    if re_rule:
        return regex.findall(re_rule, text, regex.S)

    objects = []

    if key:
        escaped_key = regex.escape(key)

        # support {}
        object_pat = rf'"{escaped_key}"\s*:\s*(\{{(?:[^{{}}]+|(?1))*\}})'
        # support []
        array_pat = rf'"{escaped_key}"\s*:\s*(\[(?:[^\[\]]+|(?1))*\])'
        # Supports basic values: string, number, null, and boolean
        literal_pat = rf'"{escaped_key}"\s*:\s*(".*?"|\d+(?:\.\d+)?|true|false|null)'

        for pat in [object_pat, array_pat, literal_pat]:
            matches = regex.findall(pat, text, regex.S | regex.I)
            for m in matches:
                try:
                    objects.append(json.loads(m))
                except Exception:
                    pass
    else:
        # If no key is provided, extract all top-level JSON objects or arrays
        block_pat = r'(\{(?:[^{}]+|(?1))*\}|\[(?:[^\[\]]+|(?1))*\])'
        matches = regex.findall(block_pat, text, regex.S)
        for m in matches:
            try:
                objects.append(json.loads(m))
            except Exception:
                pass

    return objects[0] if len(objects) == 1 else objects

class JSONExtractor:
    def __init__(self, strict_level: int = 2):
        self.strict_level = strict_level
        self.seen_json_texts = set()
        self.seen_str_json = set()
        self.seen_obj_hashes = set()

    def remove_json_comments(self, text: str) -> str:
        text = re.sub(r'//.*?(?=\n|$)', '', text)
        text = re.sub(r'/\*.*?\*/', '', text, flags=re.S)
        return text

    def hash_obj(self, obj: Any) -> str:
        try:
            dumped = json.dumps(obj, sort_keys=True, ensure_ascii=False)
            return hashlib.md5(dumped.encode('utf-8')).hexdigest()
        except Exception:
            return str(id(obj))

    def try_parse_json_recursive(self, json_str: str, max_depth: int = 5) -> Union[Dict, List, None]:
        json_str = json_str.strip()
        json_str = self.remove_json_comments(json_str)
        if not isinstance(json_str, str) or ':' not in json_str or json_str in self.seen_json_texts:
            return None
        self.seen_json_texts.add(json_str)

        parsers = []
        if self.strict_level == 2:
            parsers = [orjson.loads]
        elif self.strict_level == 1:
            parsers = [json.loads]
        elif self.strict_level == 0:
            parsers = [json.loads, json5.loads]

        for _ in range(max_depth):
            parsed = None
            for parser in parsers:
                try:
                    result = parser(json_str)
                    if isinstance(result, (dict, list)):
                        return result
                    elif isinstance(result, str):
                        json_str = result
                        break
                except Exception:
                    continue

            if parsed is None:
                unescaped = json_str.replace(r'\"', '"').replace(r'\\', '\\')
            for parser in parsers:
                try:
                    parsed = parser(unescaped)
                    break
                except Exception:
                    continue

            if parsed is None:
                return None

            if isinstance(parsed, (dict, list)):
                return parsed
            elif isinstance(parsed, str):
                json_str = parsed
                continue
        return None

    def find_key_recursively(self, obj: Any, target_key: str) -> List[Any]:
        matches = []
        if isinstance(obj, (Dict, list)):
            obj_hash = self.hash_obj(obj)
            if obj_hash in self.seen_obj_hashes:
                return matches
            self.seen_obj_hashes.add(obj_hash)

        if isinstance(obj, Dict):
            if not target_key:
                matches.append(obj)
            for k, v in obj.items():
                if k == target_key:
                    matches.append(v)
                matches.extend(self.find_key_recursively(v, target_key))
                if isinstance(v, str) and "{" in v and "}" in v:
                    parsed = self.try_parse_json_recursive(v)
                    if parsed:
                        matches.extend(self.find_key_recursively(parsed, target_key))
        elif isinstance(obj, list):
            for item in obj:
                matches.extend(self.find_key_recursively(item, target_key))
        elif isinstance(obj, str):
            if obj in self.seen_str_json:
                return matches
            self.seen_str_json.add(obj)
            parsed = self.try_parse_json_recursive(obj)
            if parsed:
                matches.extend(self.find_key_recursively(parsed, target_key))
        return matches

    def find_brace_pairs_safe(self, text: str) -> List[Any]:
        stack = []
        results = []
        for i, char in enumerate(text):
            if char == '{':
                stack.append(i)
            elif char == '}':
                if stack:
                    start = stack.pop()
                    candidate = text[start:i+1]
                    obj = self.try_parse_json_recursive(candidate)
                    if obj is not None:
                        results.append(obj)
        return results

    def extract(self, text: str, key: str = "", re_rule: str = "") -> Union[List[Union[Dict, str]], Dict, str]:
        self.seen_json_texts = set()
        self.seen_str_json = set()
        self.seen_obj_hashes = set()

        if re_rule:
            return regex.findall(re_rule, text, regex.S)

        top_obj = self.try_parse_json_recursive(text)
        if top_obj:
            matches = self.find_key_recursively(top_obj, key)
            if matches:
                return matches if len(matches) > 1 else matches[0]

        json_objects = self.find_brace_pairs_safe(text)
        if not key:
            return json_objects

        all_matches = []
        for obj in json_objects:
            all_matches.extend(self.find_key_recursively(obj, key))
        return all_matches[0] if len(all_matches) == 1 else all_matches

class JSONScanner(JSONExtractor):
    def __init__(self, strict_level: int = 2):
        super().__init__(strict_level=strict_level)

    def scan_text(self, text: str, key: str = "", re_rule: str = "") -> Union[List[Union[Dict, str]], Dict, str]:
        results = self.extract(text=text, key=key, re_rule=re_rule)
        if results:
            return results

        top_obj = self.extract(text=text)
        if not top_obj:
            return []

        stack = [top_obj] if isinstance(top_obj, (dict, list)) else []
        while stack:
            current = stack.pop()
            if isinstance(current, dict):
                for v in current.values():
                    if isinstance(v, (dict, list)):
                        stack.append(v)
                    elif isinstance(v, str):
                        inner_res = self.extract(text=v, key=key)
                        if inner_res is not None:
                            if isinstance(inner_res, list):
                                results.extend(inner_res)
                            else:
                                results.append(inner_res)
            elif isinstance(current, list):
                for item in current:
                    if isinstance(item, (dict, list)):
                        stack.append(item)
                    elif isinstance(item, str):
                        inner_res = self.extract(text=item, key=key)
                        if inner_res is not None:
                            if isinstance(inner_res, list):
                                results.extend(inner_res)
                            else:
                                results.append(inner_res)

        seen = set()
        final_results = []
        for r in results:
            if isinstance(r, str) and ':' in r and r.strip().startswith('{'):
                try:
                    if self.strict_level == 2:
                        import orjson
                        parsed_r = orjson.loads(r)
                    elif self.strict_level == 1:
                        import json
                        parsed_r = json.loads(r)
                    elif self.strict_level == 0:
                        import json5
                        parsed_r = json5.loads(r)
                    r = parsed_r
                except Exception:
                    pass

            h = str(r)
            if h not in seen:
                final_results.append(r)
                seen.add(h)
        return final_results[0] if final_results else final_results

__all__ = [
    "extract_nested_objects",
    "JSONScanner",
]

if __name__ == "__main__": 
    data = """
        <html>
            <head>...</head>
            <body>
                "{"
                <div ... class="{">
                    {
                        "a": 1,
                        "b": "2",
                        "c": [0, "3", {"_a": 4, "_b": "5"}],
                        "d": {"d0": 6, "d1": "7"},
                        "level1": {
                            "raw": "{\\"key\\": {\\"deep\\": \\"value\\"}}"
                        }
                    }
                    "{"
                    <div ... class="{">
                        {
                            "a": {"d0": 14, "d2": "15"},
                            "e": 8,
                            "f": "9",
                            "g": [10, "11", {"_a": 12, "_b": "13"}],
                            "logs": [
                                "{\\"event\\": \\"click\\", \\"meta\\": {\\"target\\": \\"button\\"}}",
                                "{\\"event\\": \\"scroll\\", \\"meta\\": {\\"target\\": \\"window\\"}}"
                            ]
                        }
                    </div>
                </div>
                {
                    "h": {"d0": 16, "d2": "17"}, // no quotes!
                    "e": 18,
                    "i": "19,
                    "j": [20, "21", {"_a": 22, "_b": "23"}],
                    "logs": [
                        "{\\"event\\": \\"click\\", \\"meta\\": {\\"target\\": \\"button\\"}}",
                        "{\\"event\\": \\"scroll\\", \\"meta\\": {\\"target\\": \\"window\\"}}"
                    ]
                }
                "}"
                {
                    "k": {"d0": 24, "d2": "25"},
                    "l": 26,
                    "m": "27,
                    "n": [28, "29", {"_a": 30, "_b": "31"}],
                    "o": '{bad: "json"}',
                "}"
            </body>
        </html>
    """

    # with open(r"", "r", encoding='utf-8') as f:
    #     text = f.read()

    print(extract_nested_objects(text=data, key="a")) # [{'d0': 14, 'd2': '15'}, 1]
    print(extract_nested_objects(text=data, key="_a")) # [4, 12, 22, 30]
    print(extract_nested_objects(text=data, key="c")) # [0, '3', {'_a': 4, '_b': '5'}]
    print(extract_nested_objects(text=data, key="e")) # [8, 18]
    print(extract_nested_objects(text=data, key="raw")) # []
    print(extract_nested_objects(text=data, key="key")) # []
    print(extract_nested_objects(text=data, key="deep")) # []
    print(extract_nested_objects(text=data, key="event")) # []
    print(extract_nested_objects(text=data, key="target")) # []
    # print(extract_nested_objects(text=text, key="emoji")) # []
    print("————————————————————————————extract_nested_objects_deep (base)————————————————————————————")
    jsonExtractor = JSONExtractor()
    jsonExtractor1 = JSONExtractor(strict_level=0)
    print(jsonExtractor.extract(text=data, key="a")) # [1, {'d0': 14, 'd2': '15'}]
    print(jsonExtractor.extract(text=data, key="_a")) # [4, 12, 22, 30]
    print(jsonExtractor.extract(text=data, key="c")) # [0, '3', {'_a': 4, '_b': '5'}]
    print(jsonExtractor1.extract(text=data, key="e")) # 8
    print(jsonExtractor.extract(text=data, key="raw")) # {"key": {"deep": "value"}}
    print(jsonExtractor.extract(text=data, key="key")) # {'deep': 'value'}
    print(jsonExtractor.extract(text=data, key="deep")) # value
    print(jsonExtractor.extract(text=data, key="event")) # ['click', 'scroll']
    print(jsonExtractor.extract(text=data, key="target")) # ['button', 'window']
    # print(jsonExtractor.extract(text=text, key="emoji")) # []
    print("————————————————————————————extract_nested_objects_deep (strong)————————————————————————————")
    jsonEScanner = JSONScanner()
    jsonEScanner1 = JSONScanner(strict_level=0)
    print(jsonEScanner.scan_text(text=data, key="a")) # [1, {'d0': 14, 'd2': '15'}]
    print(jsonEScanner.scan_text(text=data, key="_a")) # [4, 12, 22, 30]
    print(jsonEScanner.scan_text(text=data, key="c")) # [0, '3', {'_a': 4, '_b': '5'}]
    print(jsonEScanner1.scan_text(text=data, key="e")) # 8
    print(jsonEScanner.scan_text(text=data, key="raw")) # {"key": {"deep": "value"}}
    print(jsonEScanner.scan_text(text=data, key="key")) # {'deep': 'value'}
    print(jsonEScanner.scan_text(text=data, key="deep")) # value
    print(jsonEScanner.scan_text(text=data, key="event")) # ['click', 'scroll']
    print(jsonEScanner.scan_text(text=data, key="target")) # ['button', 'window']
    # print(jsonEScanner.scan_text(text=text, key="emoji")) # 😀


from scrapy_cffi.databases import RedisManager
from scrapy_cffi.spiders import Spider
from scrapy_cffi.pipelines import Pipeline

class CustomPipeline1(Pipeline):
    async def open_spider(self, spider: "Spider"):
        pass

    async def process_item(self, item: dict, spider: "Spider"):
        print(f"[CustomPipeline1] Processing item: {item}")
        item["cumtom1"] = "CustomPipeline1"
        return item

    async def close_spider(self, spider: "Spider"):
        print(f"[CustomPipeline1] {spider.__class__.name} closed.")


class CustomPipeline2(Pipeline):
    async def open_spider(self, spider: "Spider"):
        pass

    async def process_item(self, item: dict, spider: "Spider"):
        print(f"[CustomPipeline2] Processing item: {item}")
        item["cumtom2"] = "CustomPipeline2"
        return item

    async def close_spider(self, spider: "Spider"):
        print(f"[CustomPipeline2] {spider.__class__.name} closed.")


from scrapy_cffi.item import Item, Field

class CustomItem(Item):
    session_id = Field()
    ret_cookies = Field()
    session_end = Field()
    data = Field()
    cumtom1 = Field()
    cumtom2 = Field()


import argparse
from . import startproject, genspider, demo

def main():
    parser = argparse.ArgumentParser(prog="scrapy_cffi", description="scrapy_cffi CLI tool")
    subparsers = parser.add_subparsers(dest="command")

    # startproject
    sp = subparsers.add_parser("startproject", help="Create a new project")
    sp.add_argument("name", help="Project name")

    # genspider
    gp = subparsers.add_parser("genspider", help="Generate a new spider")
    gp.add_argument("-r", "--redis", action="store_true", help="Use RedisSpider")
    gp.add_argument("-m", "--rabbitmq", action="store_true", help="Use RabbitMqSpider, override -r/--redis")
    gp.add_argument("-k", "--kafka", action="store_true", help="Use Kafka Log")
    gp.add_argument("name", help="Spider name")
    gp.add_argument("domain", help="Target domain")

    # demo project
    demo_p = subparsers.add_parser("demo", help="Create a demo project")
    demo_p.add_argument("-r", "--redis", action="store_true", help="Use RedisSpider")
    demo_p.add_argument("-m", "--rabbitmq", action="store_true", help="Use RabbitMqSpider, override -r/--redis")
    demo_p.add_argument("-k", "--kafka", action="store_true", help="Use Kafka Log")

    # export
    # ep = subparsers.add_parser("export", help="Export files")
    # ep.add_argument("name", help="Filename")

    # server

    # connect

    args = parser.parse_args()

    if args.command == "startproject":
        startproject.run(args.name)
    elif args.command == "genspider":
        genspider.run(args.name, args.domain, args.redis, args.rabbitmq, args.kafka)
    # elif args.command == "export":
    #     export.run(args.name)
    elif args.command == "demo":
        result = startproject.run("demo", is_demo=True)
        if result is not None:
            return
        demo.run(args.redis, args.rabbitmq, args.kafka)
    else:
        print(f"Unknown command: {args.command}")
        parser.print_help()



import asyncio, time, inspect
from ..extensions import signals, SignalInfo
from ..utils.concurrency import safe_call
from typing import TYPE_CHECKING, Coroutine, Callable, Optional, Set
if TYPE_CHECKING:
    from ..crawler import Crawler
    from ..settings import SettingsInfo
    from ..extensions import SignalManager
    from ..mq.kafka import KafkaManager

class TaskManager:
    def __init__(
        self, 
        stop_event: asyncio.Event=None, 
        global_lock=None, 
        signalManager: "SignalManager"=None, 
        kafkaManager: "KafkaManager"=None, 
        settings: "SettingsInfo"=None, 
        is_distributed=False
    ):
        self.stop_event = stop_event
        self.global_lock = global_lock

        from ..utils import init_logger
        self.logger = init_logger(log_info=settings.LOG_INFO, logger_name=__name__)
        if kafkaManager:
            from ..utils import KafkaLoggingHandler
            kafka_handler = KafkaLoggingHandler(kafka=kafkaManager, stop_event=self.stop_event).create_fmt(settings)
            self.logger.addHandler(kafka_handler)

        self.signalManager = signalManager
        self.active_tasks = 1 if is_distributed else 0
        self.managed_tasks: Set[asyncio.Task] = set()
        self.tasks_done_event = asyncio.Event()
        self.error_event = asyncio.Event()
        self.lock = asyncio.Lock()
        self.tasks_done_event.set()

    @classmethod
    def from_crawler(cls, crawler: "Crawler", is_distributed=None):
        return cls(
            stop_event=crawler.stop_event,
            global_lock=crawler.global_lock,
            signalManager=crawler.signalManager, 
            kafkaManager=crawler.kafkaManager,
            settings=crawler.settings, 
            is_distributed=is_distributed
        )

    async def create(self, coro: Coroutine, callback: Optional[Callable] = None, **callback_kwargs):
        if self.stop_event.is_set():
            return

        async def wrapped():
            async with self.global_lock():
                try:
                    self.logger.debug(f'add task {task_id} -> {self.active_tasks}：{coro}')
                    result = await coro
                    if callback:
                        await safe_call(callback, result, **callback_kwargs)
                except (asyncio.CancelledError, KeyboardInterrupt) as e:
                    if coro and inspect.iscoroutine(coro):
                        coro.close()
                        self.error_event.set()
                    raise
                except Exception as e:
                    result = f"<Task-Error exception={repr(e)}>"
                    self.logger.error(result)
                    self.error_event.set()
                    self.signalManager.send(signal=signals.task_error, data=SignalInfo(signal_time=time.time(), reason=result))
                    raise ValueError(result)
                finally:
                    async with self.lock:
                        self.active_tasks -= 1
                        self.logger.debug(f'end task {task_id} -> {self.active_tasks}：{coro}')
                        if self.active_tasks <= 0:
                            self.tasks_done_event.set()

        loop = asyncio.get_running_loop() # Obtain the event loop here to ensure this is called within an async context
        task = loop.create_task(wrapped())
        task.add_done_callback(self.managed_tasks.discard)
        task_id = id(task)
        async with self.lock:
            self.active_tasks += 1
            self.tasks_done_event.clear()
        return task

    async def wait_until_stopped(self) -> str:
        tasks_done_task = asyncio.create_task(self.tasks_done_event.wait())
        error_task = asyncio.create_task(self.error_event.wait())
        done, pending = await asyncio.wait(
            [tasks_done_task, error_task],
            return_when=asyncio.FIRST_COMPLETED,
        )
        for task in pending:
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                pass
        return "error" if error_task in done else "tasks_done"

    def get_task_coro_path(self, task: asyncio.Task):
        try:
            coro = task.get_coro()
            if hasattr(coro, '__qualname__'):
                func_name = coro.__qualname__
            else:
                func_name = type(coro).__name__

            module = getattr(coro, '__module__', None)
            filename = getattr(coro.cr_code, 'co_filename', None) if hasattr(coro, 'cr_code') else None
            lineno = getattr(coro.cr_frame, 'f_lineno', None) if hasattr(coro, 'cr_frame') and coro.cr_frame else None

            parts = []
            if module:
                parts.append(f"{module}")
            if filename:
                parts.append(f"{filename}")
            parts.append(func_name)
            if lineno:
                parts.append(f":{lineno}")
            return " -> ".join(parts)
        except Exception as e:
            return f"<Unknown Task: {repr(e)}>"

    async def cancel_all(self):
        self.logger.info("Cancel all tasks ...")
        # current_task = asyncio.current_task()
        # all_tasks = asyncio.all_tasks()
        # cancel_targets = [t for t in all_tasks if t is not current_task and not t.done()]
        cancel_targets = [t for t in self.managed_tasks if not t.done()]
        pending_names = [self.get_task_coro_path(t) for t in cancel_targets]
        self.logger.debug(f"Cancel tasks list: {pending_names}")

        for task in cancel_targets:
            task.cancel()
        await asyncio.sleep(0)
        self.logger.info(f"Cancelled {len(cancel_targets)} coroutine tasks")
        for task in cancel_targets:
            try:
                await task
            except asyncio.CancelledError:
                pass
            except Exception as e:
                self.logger.error(f"Exception raised while cancelling task: {e}")


import shutil, toml
from pathlib import Path

def run(project_name, is_demo=False):
    base = Path(__file__).parent.parent # scrapy_cffi
    template_dir = base / "templates"
    target: Path = Path.cwd() / project_name

    if target.exists():
        print(f"Error: Project '{project_name}' already exists.")
        return False
    
    spiders_dir =  target
    shutil.copytree(template_dir / "spiders", spiders_dir)
    shutil.copytree(template_dir / "js_path", target / "js_path")

    # docker
    for docker_file in ["Dockerfile", "docker-compose.yml", ".gitignore", ".dockerignore"]:
        dockerfile_path = template_dir / "config" / docker_file
        target_docker_path = target / docker_file
        docker_code = dockerfile_path.read_text(encoding='utf-8')
        target_docker_path.write_text(docker_code, encoding='utf-8')
    
    config_data = {
        "default": {
            "project_name": project_name,
        }
    }
    config_path = target / "scrapy_cffi.toml"
    with config_path.open("w", encoding="utf-8") as f:
        toml.dump(config_data, f)
    if not is_demo:
        print(f"Project '{project_name}' created.")
        print(f"\tcd {project_name}")
        print(f"\tscrapy-cffi genspider <spider_name> <domain>")


import json
from pathlib import Path
from dotenv import dotenv_values
from typing import Type, Any, Union
try:
    from ..models.api import ComponentInfo
except BaseException as e:
    from scrapy_cffi.models.api import ComponentInfo

def _json_serialize(value: Any):
    """Convert non-serializable objects into serializable format for JSON."""
    if isinstance(value, Path):
        return str(value)
    raise TypeError(f"Type {type(value)} not JSON serializable")

# Fields that require special handling as ComponentInfo objects
_COMPONENT_FIELDS = [
    "SPIDER_INTERCEPTORS_PATH",
    "DOWNLOAD_INTERCEPTORS_PATH",
    "ITEM_PIPELINES_PATH",
    "EXTENSIONS_PATH",
]

def settings_to_env(obj: Any, env_path: Union[str, Path]):
    """
    Convert a Python object (Pydantic model or plain object) into a .env file.
    - dict/list fields are serialized as JSON
    - bool fields are converted to 'true'/'false'
    - ComponentInfo fields are written as empty JSON '{}'
    - Fields with None values are skipped
    """
    lines = []

    data = getattr(obj, "model_dump", None)
    if callable(data):
        data = obj.model_dump()
    else:
        data = obj.__dict__

    for key, value in data.items():
        if key.startswith("_") or value is None:
            continue
        # Write ComponentInfo fields as empty JSON
        if key in _COMPONENT_FIELDS:
            value_str = "{}"
        elif isinstance(value, (dict, list)):
            value_str = json.dumps(value, ensure_ascii=False, default=_json_serialize)
        elif isinstance(value, bool):
            value_str = str(value).lower()
        else:
            value_str = str(value)
        lines.append(f"{key}={value_str}")

    env_path = Path(env_path)
    env_path.parent.mkdir(parents=True, exist_ok=True)
    env_path.write_text("\n".join(lines), encoding="utf-8")


def env_to_settings(env_path: Union[str, Path], cls: Type[Any]) -> Any:
    """
    Load data from a .env file and convert it into an instance of the specified class.
    - Automatically parses JSON strings into dict/list
    - Converts 'true'/'false' to bool, numbers to int/float
    - Restores ComponentInfo fields from dict to ComponentInfo objects
    """
    raw_env = dotenv_values(env_path)
    processed = {}

    for k, v in raw_env.items():
        if v is None:
            continue
        v_strip = v.strip()
        # JSON object or list
        if (v_strip.startswith("{") and v_strip.endswith("}")) or \
           (v_strip.startswith("[") and v_strip.endswith("]")):
            try:
                processed[k] = json.loads(v_strip)
                continue
            except json.JSONDecodeError:
                pass
        # Boolean
        if v_strip.lower() in {"true", "false"}:
            processed[k] = v_strip.lower() == "true"
        # Float or int
        elif v_strip.replace('.', '', 1).replace('-', '', 1).isdigit():
            if '.' in v_strip:
                processed[k] = float(v_strip)
            else:
                processed[k] = int(v_strip)
        else:
            processed[k] = v_strip

    # Restore ComponentInfo fields
    for f in _COMPONENT_FIELDS:
        if f in processed and isinstance(processed[f], dict):
            processed[f] = ComponentInfo.from_raw(processed[f], f)

    return cls(**processed)

__all__ = [
    "settings_to_env",
    "env_to_settings"
]

if __name__ == "__main__":
    from scrapy_cffi.settings import SettingsInfo

    config = SettingsInfo()
    config.TEST_DATA = {"a":1, "b":2}
    print("Before .env:", config.TEST_DATA)

    env_path = r".env.dev"

    # Generate .env
    settings_to_env(config, env_path)

    # Load from .env during deployment
    config2 = env_to_settings(env_path, SettingsInfo)
    print("After .env:", config2.TEST_DATA)



from .base import StrictValidatedModel
from pydantic import BaseModel, Field
from typing import List, Optional
from pathlib import Path

class CPYExtension(StrictValidatedModel):
    module_name: str                   # Module name / folder name
    resource_name: Optional[str] = None  # Injected namespace name
    wrapper: Optional[str] = "wrapper.py" # Python wrapper filename
    fallback: Optional[str] = "fallback.py" # Fallback implementation filename
    build_dir: Optional[str] = "build"  # Directory for compiled shared libraries (ctypes wrapper)

class CPYExtensionsConfig(BaseModel):
    DIR: Path = Path("cpy_resources") # Root directory for all C extensions
    RESOURCES: List[CPYExtension] = Field(default_factory=list) # List of extension definitions


from .downloader import Request, HttpRequest, MediaRequest, WebSocketRequest, Response, HttpResponse, WebSocketResponse
from .sessions import CloseSignal

__all__ = [
    "Request",
    "HttpRequest",
    "MediaRequest",
    "WebSocketRequest",
    "Response",
    "HttpResponse",
    "WebSocketResponse",
    "CloseSignal"
]


from typing import TYPE_CHECKING, cast
if TYPE_CHECKING:
    from ..crawler import Crawler
    from .spiders import SpidersHooks
    from .pipelines import PipelinesHooks, _PipelinesHooks
    from .interceptors import InterceptorsHooks
    from .signals import SignalsHooks

class Hooks:
    def __init__(self, **funcs):
        for name, func in funcs.items():
            setattr(self, name, func)

def spiders_hooks(crawler: "Crawler") -> "SpidersHooks":
    hooks_obj = Hooks(
        session=Hooks(
            register_sessions=crawler.sessions.register_sessions_batch,
            get_session_cookies=crawler.sessions.get_session_cookies,
        ),
        scheduler=Hooks(
            get_start_req=getattr(crawler.scheduler, "get_start_req", lambda *a, **k: None)
        )
    )
    return cast(Hooks, hooks_obj)

def _pipelines_hooks(crawler: "Crawler") -> "_PipelinesHooks":
    hooks_obj = Hooks(
        session=Hooks(
            mark_end=crawler.sessions.mark_end,
            get_session_cookies=crawler.sessions.get_session_cookies,
        ),
    )
    return cast(Hooks, hooks_obj)

def pipelines_hooks(crawler: "Crawler") -> "PipelinesHooks":
    hooks_obj = Hooks(
        session=Hooks(
            get_session_cookies=crawler.sessions.get_session_cookies,
        ),
        signals=Hooks(
            send=crawler.signalManager.send
        )
    )
    return cast(Hooks, hooks_obj)

def interceptors_hooks(crawler: "Crawler") -> "InterceptorsHooks":
    hooks_obj = Hooks(
        session=Hooks(
            acquire=crawler.sessions.acquire,
            release=crawler.sessions.release,
            get_or_create_session=crawler.sessions.get_or_create_session,
        )
    )
    return cast(Hooks, hooks_obj)

def signals_hooks(crawler: "Crawler") -> "SignalsHooks":
    hooks_obj = Hooks(
        signals=Hooks(
            connect=crawler.signalManager.connect
        )
    )
    return cast(Hooks, hooks_obj)

__all__ = [
    "spiders_hooks",
    "_pipelines_hooks",
    "pipelines_hooks",
    "interceptors_hooks",
    "signals_hooks",
]


import logging
import sys
import os
import asyncio
import multiprocessing
from logging.handlers import TimedRotatingFileHandler, QueueHandler, QueueListener
from typing import Optional, TYPE_CHECKING
if TYPE_CHECKING:
    from ..settings import LogInfo, SettingsInfo
    from ..mq.kafka import KafkaManager

class ShortNameFormatter(logging.Formatter):
    def format(self, record):
        if '.' in record.name:
            record.name = record.name.rsplit('.', 1)[-1]
        return super().format(record)

def init_logger(log_info: "LogInfo", logger_name: str) -> logging.Logger:
    """Initialize a single process logger for the current module. Recommend passing in __name__ as logger name."""
    if not logger_name or logger_name == "root":
        raise ValueError("Please pass __name__ or module-specific logger name, not root.")

    if not log_info.LOG_ENABLED:
        logging.disable(logging.CRITICAL)
        return

    log_level = getattr(logging, log_info.LOG_LEVEL, logging.DEBUG)
    log_format = log_info.LOG_FORMAT or "%(asctime)s [%(name)s] %(levelname)s: %(message)s"
    log_dateformat = log_info.LOG_DATEFORMAT
    log_file = log_info.LOG_FILE
    log_encoding = log_info.LOG_ENCODING
    log_short_names = log_info.LOG_SHORT_NAMES
    log_formatter_cls_path = log_info.LOG_FORMATTER
    with_stream = log_info.LOG_WITH_STREAM

    if log_formatter_cls_path:
        from ..utils import load_object
        log_formatter_cls = load_object(log_formatter_cls_path)
    elif log_short_names:
        log_formatter_cls = ShortNameFormatter
    else:
        log_formatter_cls = logging.Formatter

    formatter = log_formatter_cls(fmt=log_format, datefmt=log_dateformat)

    logger = logging.getLogger(logger_name)
    logger.handlers.clear()
    # logger.propagate = False

    if with_stream:
        has_stream = any(isinstance(h, logging.StreamHandler) for h in logger.handlers)
        if not has_stream:
            stream_handler = logging.StreamHandler(sys.stdout)
            stream_handler.setLevel(log_level)
            stream_handler.setFormatter(formatter)
            logger.addHandler(stream_handler)

    if log_file:
        if not os.path.isabs(log_file):
            from . import get_run_py_dir
            base_path = get_run_py_dir()
            log_file = os.path.join(base_path, log_file)
        log_dir = os.path.dirname(log_file)
        if log_dir:
            os.makedirs(log_dir, exist_ok=True)
        has_file = any(isinstance(h, TimedRotatingFileHandler) for h in logger.handlers)
        if not has_file:
            os.makedirs(os.path.dirname(log_file), exist_ok=True)
            file_handler = TimedRotatingFileHandler(
                log_file, when='D', interval=1, backupCount=15, encoding=log_encoding
            )
            file_handler.setLevel(log_level)
            file_handler.setFormatter(formatter)
            logger.addHandler(file_handler)
    logger.setLevel(log_level)
    return logger

def start_multiprocess_log_listener(
    log_info: "LogInfo",
    with_stream: bool = True,
) -> tuple[multiprocessing.Queue, QueueListener]:
    log_level = getattr(logging, log_info.LOG_LEVEL, logging.DEBUG)
    log_format = log_info.LOG_FORMAT or "%(asctime)s [%(name)s] %(levelname)s: %(message)s"
    log_dateformat = log_info.LOG_DATEFORMAT
    log_file = log_info.LOG_FILE
    log_encoding = log_info.LOG_ENCODING
    log_short_names = log_info.LOG_SHORT_NAMES
    log_formatter_cls_path = log_info.LOG_FORMATTER

    if log_formatter_cls_path:
        from ..utils import load_object
        log_formatter_cls = load_object(log_formatter_cls_path)
    elif log_short_names:
        log_formatter_cls = ShortNameFormatter
    else:
        log_formatter_cls = logging.Formatter

    formatter = log_formatter_cls(fmt=log_format, datefmt=log_dateformat)

    handlers = []

    if with_stream:
        stream_handler = logging.StreamHandler(sys.stdout)
        stream_handler.setLevel(log_level)
        stream_handler.setFormatter(formatter)
        handlers.append(stream_handler)

    if log_file:
        os.makedirs(os.path.dirname(log_file), exist_ok=True)
        file_handler = TimedRotatingFileHandler(
            log_file, when='D', interval=1, backupCount=15, encoding=log_encoding
        )
        file_handler.setLevel(log_level)
        file_handler.setFormatter(formatter)
        handlers.append(file_handler)

    log_queue = multiprocessing.Queue()

    root_logger = logging.getLogger()
    root_logger.setLevel(log_level)
    root_logger.handlers.clear()
    root_logger.addHandler(QueueHandler(log_queue))

    listener = QueueListener(log_queue, *handlers)
    listener.start()

    return log_queue, listener

def init_logger_multiprocessing(
    logger_name: str, 
    log_level="DEBUG", 
    log_queue: Optional[multiprocessing.Queue] = None, 
    with_stream=True,
    formatter: Optional[logging.Formatter] = None,
    extra_handlers=None
) -> logging.Logger:
    logger = logging.getLogger(logger_name)
    logger.setLevel(getattr(logging, log_level.upper(), logging.DEBUG))
    logger.handlers.clear()
    logger.propagate = False
    if formatter is None:
        formatter = logging.Formatter('%(asctime)s - %(filename)s:%(lineno)d - %(levelname)s - %(message)s')
    if with_stream:
        stream_handler = logging.StreamHandler()
        stream_handler.setLevel(logging.NOTSET)
        stream_handler.setFormatter(formatter)
        logger.addHandler(stream_handler)
    if log_queue:
        queue_handler = QueueHandler(log_queue)
        logger.addHandler(queue_handler)
    if extra_handlers:
        for h in extra_handlers:
            logger.addHandler(h)
    return logger

class KafkaLoggingHandler(logging.Handler):
    def __init__(self, kafka: "KafkaManager", topic: str = "scrapy_cffi", stop_event: asyncio.Event = None):
        super().__init__()
        self.kafka = kafka
        self.topic = topic
        self.stop_event = stop_event
        self.queue = asyncio.Queue()
        self.task = asyncio.create_task(self._sender_loop())
        self._closed = False

    def create_fmt(self, settings: "SettingsInfo") -> "KafkaLoggingHandler":
        self.setFormatter(logging.Formatter(settings.LOG_INFO.LOG_FORMAT, datefmt=settings.LOG_INFO.LOG_DATEFORMAT))
        return self
    
    async def _sender_loop(self):
        try:
            while not self.stop_event.is_set() or not self.queue.empty():
                try:
                    msg = await asyncio.wait_for(self.queue.get(), timeout=1)
                    try:
                        await self.kafka.produce(self.topic, msg)
                    finally:
                        self.queue.task_done()
                except asyncio.TimeoutError:
                    continue
        except asyncio.CancelledError:
            # Drain remaining messages before exiting
            while not self.queue.empty():
                msg = await self.queue.get()
                try:
                    await self.kafka.produce(self.topic, msg)
                finally:
                    self.queue.task_done()
                    
                # try:
                #     msg = await asyncio.wait_for(self.queue.get(), timeout=3.0)
                #     try:
                #         await self.kafka.produce(self.topic, msg)
                #     finally:
                #         self.queue.task_done()
                # except asyncio.TimeoutError:
                #     pass
            raise
        except Exception as e:
            print(f"Kafka send failed: {e}")

    def emit(self, record: logging.LogRecord):
        try:
            msg = self.format(record).encode("utf-8")
            self.queue.put_nowait(msg)
        except Exception as e:
            print(f"KafkaLoggingHandler emit error: {e}")

    def close(self):
        if self._closed:
            return
        self._closed = True
        self.stop_event.set()

        if self.task and not self.task.done():
            loop = asyncio.get_running_loop()
            # Run the sender task to completion synchronously
            future = asyncio.run_coroutine_threadsafe(self.task, loop)
            try:
                future.result()  # Block until the task completes
            except Exception as e:
                print(f"KafkaLoggingHandler close error: {e}")
        super().close()


# Core Component Signals
engine_started = object() # Engine started          SignalInfo(signal_time=time.time())
engine_stopped = object() # Engine stopped          SignalInfo(signal_time=time.time())
scheduler_empty = object() # Scheduler is empty     SignalInfo(signal_time=time.time())
task_error = object() # Task failed                 SignalInfo(signal_time=time.time(), reason=result)

# Spider Lifecycle Signals
spider_opened = object() # Spider opened     SignalInfo(spider: Spider, signal_time=time.time())
spider_closed = object() # Spider closed     SignalInfo(spider: Spider, signal_time=time.time())
spider_error = object() # Spider error       SignalInfo(response: Response, exception: BaseException, spider: Spider, signal_time=time.time())

# Request Scheduling Signals
request_scheduled = object() # Request successfully scheduled   SignalInfo(signal_time=time.time(), request=request)
request_dropped = object() # Request was dropped                SignalInfo(signal_time=time.time(), request=request, reason: str)

# Downloader Signals
request_reached_downloader = object() # Request sent to downloader  SignalInfo(signal_time=time.time(), request=request)
response_received = object() # Response received                    SignalInfo(signal_time=time.time(), request=request, response=response)

# Item Pipeline Signals
item_scraped = object() # Item scraped successfully         SignalInfo(signal_time=time.time(), item: Item, spider: Spider)
item_dropped = object() # Item was dropped                  SignalInfo(signal_time=time.time(), item: Item, reason: str)
item_error = object() # Exception during item processing    SignalInfo(signal_time=time.time(), item: Item, exception: BaseException)


import importlib, importlib.util, sys, os, inspect, json, traceback, toml
import asyncio
from contextlib import asynccontextmanager
from typing import Any, Callable, TYPE_CHECKING, Union, Dict, List, Optional, Type
if TYPE_CHECKING:
    from logging import Logger

def get_run_py_dir():
    from pathlib import Path
    return Path(sys.argv[0]).resolve().parent

def get_or_create_loop():
    if sys.platform.startswith("win"):
        asyncio.set_event_loop_policy(asyncio.WindowsProactorEventLoopPolicy())

    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        try:
            loop = asyncio.get_event_loop()
            if loop.is_closed():
                raise RuntimeError
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
    return loop

def setup_uvloop_once():
    if getattr(setup_uvloop_once, "_done", False):
        return
    try:
        if sys.platform != "win32":
            import uvloop
            asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
            print("uvloop enabled")
        else:
            print("uvloop not available on Windows")
    except ImportError as e:
        print("uvloop is not install")
    finally:
        setup_uvloop_once._done = True

def async_context_factory(
    max_tasks: Optional[int] = None,
    semaphore_cls: Union[Type[asyncio.Semaphore], None] = None
):
    if max_tasks is None:
        @asynccontextmanager
        async def empty_context():
            yield
        return empty_context
    else:
        sem_cls = semaphore_cls or asyncio.BoundedSemaphore
        sem = sem_cls(max_tasks)

        @asynccontextmanager
        async def sem_context():
            async with sem:
                yield
        return sem_context

class ResultHolder:
    def __init__(self):
        loop = get_or_create_loop()
        self._future = loop.create_future()

    def set_result(self, value):
        if not self._future.done():
            self._future.set_result(value)

    async def get_result(self):
        return await self._future

async def cancel_all_tasks(timeout: float = 5.0):
    current_task = asyncio.current_task()
    all_tasks = asyncio.all_tasks()
    cancel_targets = [t for t in all_tasks if t is not current_task and not t.done()]
    print(f"[cancel] Found {len(cancel_targets)} tasks to cancel")
    for task in cancel_targets:
        task.cancel()

    await asyncio.sleep(0)

    try:
        await asyncio.wait(cancel_targets, timeout=timeout)
    except Exception as e:
        print(f"[cancel] Error while waiting for task cancellation: {e}")

    still_pending = [t for t in cancel_targets if not t.done()]
    if still_pending:
        print(f"[cancel] {len(still_pending)} tasks were not cancelled within {timeout}s, calling os._exit(0) for forced termination")
        for task in still_pending:
            print(f"\n--- Incomplete task: {task.get_name()} ---")
            for frame in task.get_stack():
                print("".join(traceback.format_stack(f=frame)))
        await asyncio.sleep(1)
        os._exit(0)
    else:
        print(f"[cancel] All tasks successfully cancelled within {timeout}s")

    for task in cancel_targets:
        try:
            await asyncio.wait_for(task, timeout=1)
        except asyncio.CancelledError:
            pass
        except asyncio.TimeoutError:
            pass

# Dynamically import class by its full path and return the class object
def load_object(path: str, base_module: str = None):
    if not path:
        raise ValueError("Empty path is not allowed")

    if '.' not in path:
        # Short class name, try loading from base_module or caller's module
        if base_module:
            module = importlib.import_module(base_module)
        else:
            # Get caller's module (2nd frame on the stack)
            frame = sys._getframe(1)
            caller_globals = frame.f_globals
            module_name = caller_globals.get('__name__', '__main__')
            module = sys.modules[module_name]
        try:
            return getattr(module, path)
        except AttributeError as e:
            raise AttributeError(f"Class '{path}' not found in module '{module.__name__}'") from e
    else:
        try:
            module_path, class_name = path.rsplit('.', 1)
            module = importlib.import_module(module_path)
            return getattr(module, class_name)
        except (ImportError, AttributeError, ValueError) as e:
            raise ImportError(f"Cannot load '{path}': {e}") from e

# ——————————————————————————————————————————————————————————————————————————————————————————————————————————————————————————————————————————————————————
# Convert SettingsInfo (Pydantic model) to scrapy-style settings.py format
def to_scrapy_settings_py(settings_obj) -> str:
    from pydantic import BaseModel
    def serialize_value(value: Any) -> str:
        if isinstance(value, BaseModel):
            return json.dumps(value.model_dump(), indent=4, ensure_ascii=False)
        elif isinstance(value, dict):
            return json.dumps(value, indent=4, ensure_ascii=False)
        elif isinstance(value, (list, tuple)):
            return repr(value)
        elif isinstance(value, str):
            return repr(value)
        else:
            return str(value)

    lines = []
    for field_name, value in settings_obj.model_dump().items():
        if value is not None:
            line = f"{field_name} = {serialize_value(value)}"
            lines.append(line)
    return "\n".join(lines)

# Load settings from a given path (supports Python or JSON)
def load_settings_with_path(settings_path: str=""):
    from ..settings import SettingsInfo
    if settings_path == "":
        settings_path = str(get_run_py_dir() / "settings.py")
    if ".py" in settings_path:
        custom_settings = load_settings_from_py(settings_path)
    else:
        # Try parsing as JSON if not a Python file
        try:
            with open(settings_path, 'r', encoding='utf-8') as f:
                custom_settings: Dict[str, Union[List, str, None]] = json.loads(f.read())
                custom_settings = {k.upper(): v for k, v in custom_settings.items()}
        except Exception as e:
            return f"Error parsing JSON config: {e}"
    return SettingsInfo(**custom_settings)

# Load settings from a Python file as a dictionary
def load_settings_from_py(filepath: str, auto_upper=True) -> Dict[str, Any]:
    if not os.path.exists(filepath):
        raise FileNotFoundError(f"Settings file not found: {filepath}")
    
    module_name = "__user_settings__"
    spec = importlib.util.spec_from_file_location(module_name, filepath)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    settings = {
        (key.upper() if auto_upper else key): getattr(module, key)
        for key in dir(module)
        if not key.startswith("__")
    }
    return settings

# Convert a Python settings file to a .toml file
def convert_to_toml(py_path: str, toml_path: str):
    config = load_settings_from_py(py_path, auto_upper=False)
    toml_dict = {}
    for key, value in config.items():
        if isinstance(value, dict):
            toml_dict[key] = value
        else:
            toml_dict[key] = value
    with open(toml_path, "w", encoding="utf-8") as f:
        toml.dump(toml_dict, f)
    print(f"Converted {py_path} to {toml_path} successfully.")

# ——————————————————————————————————————————————————————————————————————————————————————————————————————————————————————————————————————————————————————
def get_class_name(item):
    cls = item if isinstance(item, type) else item.__class__
    return f"{cls.__module__}.{cls.__name__}"

# Get the spider class
def get_all_spiders_cls(spiders_dir: str):
    from ..spiders import BaseSpider
    if not os.path.isabs(spiders_dir):
        spiders_dir = os.path.join(get_run_py_dir(), spiders_dir)
    spider_classes = []
    for root, dirs, files in os.walk(spiders_dir):
        dirs[:] = [d for d in dirs if d != '__pycache__']
        for file in files:
            if not file.endswith('.py'):
                continue
            if file.startswith('__') or file.startswith('test_'):
                continue

            full_path = os.path.join(root, file)
            module_name = os.path.splitext(os.path.relpath(full_path, spiders_dir))[0].replace(os.path.sep, '.')
            
            spec = importlib.util.spec_from_file_location(module_name, full_path)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)

            for obj in vars(module).values():
                if inspect.isclass(obj) and issubclass(obj, BaseSpider) and obj.__module__ == module.__name__:
                    spider_classes.append(obj)
    return spider_classes

# Get all spider names
def get_all_spiders_name(logger: "Logger"=None, spiders_cls_list=None):
    spiders_name = [spider.name for spider in spiders_cls_list]
    logger.debug(f"all_spiders：{spiders_name}")
    return spiders_name
    
async def run_with_timeout(
    func: Callable[..., Any],
    *args,
    stop_event: asyncio.Event,
    timeout: float = 1.0,
    max_total_time: Optional[float] = None,
    **kwargs
) -> Any:
    is_async = inspect.iscoroutinefunction(func)
    start_time = asyncio.get_running_loop().time()

    while not stop_event.is_set():
        try:
            if is_async:
                task = asyncio.create_task(func(*args, **kwargs))
            else:
                task = asyncio.create_task(asyncio.to_thread(func, *args, **kwargs))
            return await asyncio.wait_for(task, timeout=timeout)
        except asyncio.TimeoutError:
            if not task.done():
                task.cancel()
                try:
                    await task
                except asyncio.CancelledError:
                    pass

            now = asyncio.get_running_loop().time()
            if max_total_time is not None and (now - start_time > max_total_time):
                raise asyncio.TimeoutError("Maximum total wait time exceeded")

            await asyncio.sleep(0.05)
        except BaseException as e:
            if stop_event.is_set():
                raise asyncio.CancelledError("Stopped by stop_event")
            raise
    raise asyncio.CancelledError("stop_event set during call")


from scrapy_cffi.interceptors import DownloadInterceptor
from scrapy_cffi.internet import *
from typing import Union, TYPE_CHECKING
if TYPE_CHECKING:
    from scrapy_cffi.spiders import BaseSpider

class CustomDownloadInterceptor1(DownloadInterceptor):
    RequestType = Union[HttpRequest, WebSocketRequest]
    ResponseType = Union[HttpResponse, WebSocketResponse]

    async def request_intercept(self, request: RequestType, spider: "BaseSpider"):
        return None

    async def response_intercept(self, request: RequestType, response: ResponseType, spider: "BaseSpider"):
        return response
    
    async def exception_intercept(self, request: RequestType, exception: BaseException, spider: "BaseSpider"):
        return exception
    
class CustomDownloadInterceptor2(DownloadInterceptor):
    RequestType = Union[HttpRequest, WebSocketRequest]
    ResponseType = Union[HttpResponse, WebSocketResponse]

    async def request_intercept(self, request: RequestType, spider: "BaseSpider"):
        return None

    async def response_intercept(self, request: RequestType, response: ResponseType, spider: "BaseSpider"):
        return response
    
    async def exception_intercept(self, request: RequestType, exception: BaseException, spider: "BaseSpider"):
        return exception


import asyncio
import inspect
import random
from functools import wraps
from typing import TYPE_CHECKING, Union, List, Dict, Optional, Callable, Tuple

try:
    from aiokafka import AIOKafkaProducer, AIOKafkaConsumer
    from aiokafka.admin import AIOKafkaAdminClient, NewTopic
    from aiokafka.errors import KafkaConnectionError
except ImportError as e:
    raise ImportError(
        "Missing aiokafka dependencies. Please install: pip install aiokafka"
    ) from e

from tenacity import retry, wait_fixed, retry_if_exception_type

if TYPE_CHECKING:
    from ..crawler import Crawler

def auto_retry(func):
    @wraps(func)
    @retry(
        wait=wait_fixed(1),
        retry=retry_if_exception_type(KafkaConnectionError),
        reraise=True
    )
    async def wrapper(self, *args, **kwargs):
        if self.stop_event.is_set():
            raise asyncio.CancelledError("Stop event set, abort Kafka operation")
        try:
            return await func(self, *args, **kwargs)
        except KafkaConnectionError:
            if self.stop_event.is_set():
                raise asyncio.CancelledError("Stop event set during reconnect")
            await self.connect()
            return await func(self, *args, **kwargs)
    return wrapper

class KafkaManager:
    def __init__(
        self,
        stop_event: asyncio.Event = None,
        kafka_url: Union[str, List[str]] = None,
        loop: asyncio.AbstractEventLoop = None,
        consumer_group: str = "scrapy_cffi",
        persistent_time: int = 7*24*60*60*1000
    ):
        self.stop_event = stop_event or asyncio.Event()
        self.loop = loop or asyncio.get_running_loop()
        self.consumer_group = consumer_group
        self.persistent_time = persistent_time

        if isinstance(kafka_url, str):
            self.mq_mode = "single"
            self._nodes = [kafka_url]
        elif isinstance(kafka_url, list):
            self.mq_mode = "cluster"
            if not kafka_url:
                raise ValueError("Empty Kafka cluster node list")
            self._nodes = kafka_url
        else:
            raise ValueError("kafka_url must be str or list of str")

        self._bootstrap_servers: str = None
        self._producer: Optional[AIOKafkaProducer] = None
        self._consumers: Dict[Tuple[str, str], AIOKafkaConsumer] = {}
        self._consumer_tasks: List[asyncio.Task] = []
        self._callbacks: Dict[Tuple[str, str], Callable[[bytes], None]] = {}
        self._method_cache: Dict[str, callable] = {}

    @classmethod
    def from_crawler(cls, crawler: "Crawler"):
        return cls(
            stop_event=crawler.stop_event,
            kafka_url=crawler.settings.KAFKA_INFO.resolved_url,
            consumer_group=crawler.settings.KAFKA_INFO.CONSUMER_GROUP,
            persistent_time=crawler.settings.KAFKA_INFO.PERSISTENT_TIME,
        )

    @auto_retry
    async def connect(self):
        self._bootstrap_servers = random.choice(self._nodes)
        if self._producer is None:
            self._producer = AIOKafkaProducer(
                loop=self.loop,
                bootstrap_servers=self._bootstrap_servers,
            )
            await self._producer.start()

    def __getattribute__(self, name: str):
        if name.startswith("_") or name in ("_method_cache", "connect", "close"):
            return super().__getattribute__(name)

        attr = super().__getattribute__(name)
        if not callable(attr) or not inspect.iscoroutinefunction(attr):
            return attr

        method_cache = super().__getattribute__("_method_cache")
        if name not in method_cache:
            @wraps(attr)
            async def wrapper(*args, **kwargs):
                if self.stop_event.is_set():
                    raise asyncio.CancelledError(f"Stop event set, abort Kafka operation: {name}")
                try:
                    return await attr(*args, **kwargs)
                except KafkaConnectionError:
                    await self.connect()
                    return await attr(*args, **kwargs)
            method_cache[name] = wrapper
        return method_cache[name]

    @auto_retry
    async def ensure_topic(self, topic: str, num_partitions: int = 1, replication_factor: int = 1):
        admin = AIOKafkaAdminClient(
            bootstrap_servers=self._bootstrap_servers,
            loop=self.loop
        )
        await admin.start()
        try:
            existing = await admin.list_topics()
            if topic not in existing:
                new_topic = NewTopic(
                    name=topic,
                    num_partitions=num_partitions,
                    replication_factor=replication_factor,
                    topic_configs={
                        "retention.ms": str(self.persistent_time),
                        "cleanup.policy": "delete"
                    }
                )
                await admin.create_topics([new_topic])
        finally:
            await admin.close()

    @auto_retry
    async def produce(self, topic: str, message: bytes, key: bytes = None):
        if self.stop_event.is_set():
            raise asyncio.CancelledError("Stop event set, abort Kafka produce")
        if self._producer is None:
            await self.connect()
        res = await self._producer.send_and_wait(topic, message, key=key)
        return res

    async def produce_async(self, topic: str, message: bytes, key: bytes = None):
        if self.stop_event.is_set():
            return
        if self._producer is None:
            await self.connect()
        result = self._producer.send(topic, message, key=key)
        if inspect.iscoroutine(result):
            await result
        else:
            await asyncio.to_thread(lambda: result)

    async def ensure_topics(self, topics: List[str]):
        for t in topics:
            await self.ensure_topic(t)

    @auto_retry
    async def _consume_loop(self, topic: str, consumer_group: str):
        key = (topic, consumer_group)
        consumer = self._consumers[key]
        callback = self._callbacks[key]

        try:
            while not self.stop_event.is_set():
                try:
                    msg = await asyncio.wait_for(consumer.getone(), timeout=1.0)
                    if inspect.iscoroutinefunction(callback):
                        await callback(msg.value)
                    else:
                        callback(msg.value)
                except asyncio.TimeoutError:
                    continue
        except asyncio.CancelledError:
            pass

    @auto_retry
    async def register_consumer(
        self,
        topic: str,
        callback: Callable[[bytes], None],
        consumer_group: Optional[str] = None,
        auto_offset_reset: str = "earliest"
    ):
        consumer_group = consumer_group or self.consumer_group
        key = (topic, consumer_group)

        if key not in self._consumers:
            if self._producer is None:
                await self.connect()
            consumer = AIOKafkaConsumer(
                topic,
                loop=self.loop,
                bootstrap_servers=self._bootstrap_servers,
                group_id=consumer_group,
                enable_auto_commit=True,
                auto_offset_reset=auto_offset_reset
            )
            await consumer.start()
            self._consumers[key] = consumer
            self._callbacks[key] = callback

            task = self.loop.create_task(self._consume_loop(topic, consumer_group))
            self._consumer_tasks.append(task)

    async def close(self):
        for task in self._consumer_tasks:
            task.cancel()
        await asyncio.gather(*self._consumer_tasks, return_exceptions=True)

        await asyncio.gather(*[c.stop() for c in self._consumers.values()], return_exceptions=True)
        self._consumers.clear()
        self._callbacks.clear()

        if self._producer:
            await self._producer.stop()


from typing import Protocol, Dict, TYPE_CHECKING
if TYPE_CHECKING:
    from ..core.sessions import SessionWrapper

class SessionHooks(Protocol):
    def acquire(self, session_id: str) -> None: ...

    def release(self, session_id: str) -> None: ...

    def get_or_create_session(self, session_id: str, cookies: Dict=None) -> "SessionWrapper": ...

class InterceptorsHooks(Protocol):
    session: SessionHooks



import sys, warnings
from pathlib import Path
import importlib.util
from types import ModuleType
from typing import Optional, TYPE_CHECKING, List
if TYPE_CHECKING:
    from ..models.api import CPYExtension

class CExtensionLoader:
    def __init__(self, resource_dir: Optional[Path] = None):
        self.user_base = (Path(sys.argv[0]).parent / resource_dir) if resource_dir else None
        self.framework_base = Path(__file__).parent / "cpy_resources"

    def _load_py_module(self, path: Path, module_name: str) -> ModuleType:
        spec = importlib.util.spec_from_file_location(module_name, path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        return module

    def _load_ctypes_wrapper(self, wrapper_path: Path) -> ModuleType:
        # Wrapper. py is responsible for loading ctypes dynamic libraries internally
        spec = importlib.util.spec_from_file_location(wrapper_path.stem, wrapper_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        return module

    def _try_load_path(self, path: Path, cfg: "CPYExtension") -> Optional[ModuleType]:
        # 1. wrapper.py
        wrapper_path = path / (cfg.wrapper or "wrapper.py")
        if wrapper_path.exists():
            try:
                return self._load_ctypes_wrapper(wrapper_path)
            except Exception as e:
                pass

        # 2. fallback.py
        fallback_path = path / (cfg.fallback or "fallback.py")
        if fallback_path.exists():
            try:
                return self._load_py_module(fallback_path, cfg.module_name)
            except Exception as e:
                pass

        return None

    def load_module(self, cfg: "CPYExtension") -> Optional[ModuleType]:
        paths_to_try = []
        module_folder = cfg.module_name

        if self.user_base:
            paths_to_try.append(self.user_base / module_folder)
        if self.framework_base:
            paths_to_try.append(self.framework_base / module_folder)

        for path in paths_to_try:
            module = self._try_load_path(path, cfg)
            if module:
                return module

        warnings.warn(f"Cannot load module {cfg.module_name} from user or framework directories.")
        return None

    def load_all(self, configs: List["CPYExtension"], inject_globals=True) -> dict[str, Optional[ModuleType]]:
        loaded = {}
        for cfg in configs:
            mod = self.load_module(cfg)
            name = cfg.resource_name or cfg.module_name
            loaded[name] = mod
            if inject_globals and mod:
                globals()[name] = mod
                sys.modules[name] = mod
        return loaded



import gzip
from . import blackboxprotobuf
from typing import Union, Dict, List, Tuple

class ProtobufFactory(object):
    @staticmethod
    def protobuf_encode(data: Dict, typedef: Dict) -> bytes:
        return blackboxprotobuf.encode_message(data, typedef)

    @staticmethod
    def protobuf_decode(data: bytes) -> Tuple[Dict, Dict]:
        data, typedef = blackboxprotobuf.decode_message(data)
        return data, typedef

    @staticmethod
    def encode_message_length(length: int) -> bytes:
        if not (0 <= length <= 0xFFFFFFFF):
            raise ValueError("Message length must be between 0 and 2^32-1")
        return length.to_bytes(4, byteorder="big")

    @staticmethod
    def decode_message_length(data: bytes) -> int:
        if len(data) != 4:
            raise ValueError("Expected 4 bytes for message length")
        return int.from_bytes(data, byteorder="big")

    @staticmethod
    def grpc_encode(data: Dict, typedef: Dict, is_gzip: bool=False) -> bytes:
        encode_data = ProtobufFactory.protobuf_encode(data, typedef)
        if is_gzip:
            encode_data = gzip.compress(encode_data)
            return bytes([1]) + ProtobufFactory.encode_message_length(len(encode_data)) + encode_data
        return bytes([0]) + ProtobufFactory.encode_message_length(len(encode_data)) + encode_data
    
    @staticmethod
    def grpc_stream_encode(data: List[Tuple[Dict, Dict]], is_gzip=False) -> bytes:
        stream_bytes = b""
        for data, typedef in data:
            encoded_msg = ProtobufFactory.grpc_encode(data, typedef, is_gzip)
            stream_bytes += encoded_msg
        return stream_bytes
        
    @staticmethod
    def grpc_decode(data: bytes) -> Union[Tuple[Dict, Dict], List[Tuple[Dict, Dict]]]:
        results = []
        offset = 0
        total_len = len(data)

        while offset < total_len:
            if total_len - offset < 5:
                raise ValueError("Incomplete grpc message header")

            compress_flag = data[offset]
            offset += 1

            msg_length_bytes = data[offset:offset+4]
            if len(msg_length_bytes) < 4:
                raise ValueError("Incomplete grpc message length")
            msg_length = ProtobufFactory.decode_message_length(msg_length_bytes)
            offset += 4

            if total_len - offset < msg_length:
                raise ValueError("Incomplete grpc message body")
            message_data = data[offset:offset+msg_length]
            offset += msg_length
            if compress_flag == 1:
                decompressed = gzip.decompress(message_data)
                decoded_data, typedef = ProtobufFactory.protobuf_decode(decompressed)
            else:
                decoded_data, typedef = ProtobufFactory.protobuf_decode(message_data)
            results.append((decoded_data, typedef))
        return results[0] if len(results) == 1 else results
    
__all__ = [
    "ProtobufFactory"
]


import asyncio
try:
    import websockets
    from websockets.exceptions import ConnectionClosed
except ImportError as e:
    raise ImportError(
        "Missing websockets dependencies. (dev env websockets==15.0.1)"
        "Please install: pip install websockets"
    ) from e
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from websockets.asyncio.server import ServerConnection

ROUTES: dict[str, callable] = {}

def route(path):
    def decorator(func):
        ROUTES[path] = func
        return func
    return decorator

@route("/")
async def root_handler(websocket: "ServerConnection"):
    print(f"[/] new connected: {websocket.remote_address}")
    await websocket.send("welcome to /")
    async for message in websocket:
        print(f"[/] received: {message.decode()}")
        await websocket.send(f"[/] send: {message.decode()}")

@route("/hello")
async def hello_handler(websocket: "ServerConnection"):
    print(f"[/hello] new connected: {websocket.remote_address}")
    await websocket.send("Hi from /hello!")
    async for message in websocket:
        print(f"[/hello] received: {message.decode()}")
        await websocket.send(f"[/hello] send: {message.decode()}")

async def handle_connection(websocket: "ServerConnection"):
    path = websocket.request.path
    handler = ROUTES.get(path)
    if handler:
        try:
            await handler(websocket)
        except ConnectionClosed as e:
            print(f"[{path}] closed: {e}")
        except Exception as e:
            print(f"[{path}] error: {e}")
    else:
        await websocket.send("Unsupported path")
        await websocket.close()

async def main():
    async with websockets.serve(handle_connection, "localhost", 8765):
        print("WebSocket server started on ws://localhost:8765")
        await asyncio.Future()  # run forever

if __name__ == "__main__":
    asyncio.run(main())

"""
import asyncio
try:
    import websockets
    from websockets import WebSocketServerProtocol
except ImportError as e:
    raise ImportError(
        "Missing websockets dependencies. (dev env websockets==13.1)"
        "Please install: pip install websockets"
    ) from e

ROUTES: dict[str, callable] = {}

def route(path):
    def decorator(func):
        ROUTES[path] = func
        return func
    return decorator

@route("/")
async def root_handler(websocket: WebSocketServerProtocol, path: str):
    print(f"[/] new connected: {websocket.remote_address}")
    await websocket.send("webcome to /")
    async for message in websocket:
        print(f"[/] received: {message.decode()}")
        await websocket.send(f"[/] send: {message.decode()}")

@route("/hello")
async def hello_handler(websocket: WebSocketServerProtocol, path: str):
    print(f"[/hello] new connected: {websocket.remote_address}")
    await websocket.send("Hi from /hello!")
    async for message in websocket:
        print(f"[/hello] received: {message.decode()}")
        await websocket.send(f"[/hello] send: {message.decode()}")

async def handle_connection(websocket: WebSocketServerProtocol, path: str):
    handler = ROUTES.get(path)
    if handler:
        try:
            await handler(websocket, path)
        except websockets.exceptions.ConnectionClosed as e:
            print(f"[{path}] close: {e}")
        except Exception as e:
            print(f"[{path}] error: {e}")
    else:
        await websocket.send("Unsupported path")
        await websocket.close()

async def main():
    server = await websockets.serve(
        handle_connection,
        host="localhost",
        port=8765,
        subprotocols=["chat"]
    )
    print("WebSocket server started on ws://localhost:8765")
    await server.wait_closed()

if __name__ == "__main__":
    asyncio.run(main())
"""